import os
import warnings
from copy import deepcopy
from glob import glob
from socket import getfqdn

from django.utils.log import DEFAULT_LOGGING
from django.utils.translation import gettext_lazy as _

from dynaconf import LazySettings

from .util.core_helpers import (
    get_app_packages,
    get_app_settings_overrides,
    merge_app_settings,
    monkey_patch,
)

monkey_patch()

IN_PYTEST = "PYTEST_CURRENT_TEST" in os.environ or "TOX_ENV_DIR" in os.environ
PYTEST_SETUP_DATABASES = [("default", "default_oot")]

ENVVAR_PREFIX_FOR_DYNACONF = "ALEKSIS"
DIRS_FOR_DYNACONF = ["/etc/aleksis"]
MERGE_ENABLED_FOR_DYNACONF = True

SETTINGS_FILE_FOR_DYNACONF = []
for directory in DIRS_FOR_DYNACONF:
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*.json"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*.ini"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*.yaml"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*.toml"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*/*.json"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*/*.ini"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*/*.yaml"))
    SETTINGS_FILE_FOR_DYNACONF += glob(os.path.join(directory, "*/*.toml"))

_settings = LazySettings(
    ENVVAR_PREFIX_FOR_DYNACONF=ENVVAR_PREFIX_FOR_DYNACONF,
    SETTINGS_FILE_FOR_DYNACONF=SETTINGS_FILE_FOR_DYNACONF,
    MERGE_ENABLED_FOR_DYNACONF=MERGE_ENABLED_FOR_DYNACONF,
)

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

# Cache directory for external operations
CACHE_DIR = _settings.get("caching.dir", os.path.join(BASE_DIR, "cache"))

SILENCED_SYSTEM_CHECKS = []

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = _settings.get("secret_key", "DoNotUseInProduction")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = _settings.get("maintenance.debug", False)
INTERNAL_IPS = _settings.get("maintenance.internal_ips", [])

UWSGI = {
    "module": "aleksis.core.wsgi",
}
UWSGI_SERVE_STATIC = True
UWSGI_SERVE_MEDIA = False

DEV_SERVER_PORT = 8000
DJANGO_VITE_DEV_SERVER_PORT = DEV_SERVER_PORT + 1

ALLOWED_HOSTS = _settings.get("http.allowed_hosts", [getfqdn(), "localhost", "127.0.0.1", "[::1]"])
BASE_URL = _settings.get(
    "http.base_url",
    f"http://localhost:{DEV_SERVER_PORT}" if DEBUG else f"https://{ALLOWED_HOSTS[0]}",
)


def generate_trusted_origins():
    origins = []
    origins += [f"http://{host}" for host in ALLOWED_HOSTS]
    origins += [f"https://{host}" for host in ALLOWED_HOSTS]
    if DEBUG:
        origins += [f"http://{host}:{DEV_SERVER_PORT}" for host in ALLOWED_HOSTS]
        origins += [f"http://{host}:{DJANGO_VITE_DEV_SERVER_PORT}" for host in ALLOWED_HOSTS]
    return origins


CSRF_TRUSTED_ORIGINS = _settings.get("http.trusted_origins", generate_trusted_origins())

# Application definition
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.humanize",
    "django_uwsgi",
    "django_extensions",
    "guardian",
    "rules.apps.AutodiscoverRulesConfig",
    "haystack",
    "polymorphic",
    "dj_cleavejs.apps.DjCleaveJSConfig",
    "dbbackup",
    "django_celery_beat",
    "django_celery_results",
    "celery_progress",
    "health_check.contrib.celery",
    "djcelery_email",
    "celery_haystack",
    "sass_processor",
    "django_any_js",
    "django_yarnpkg",
    "django_vite",
    "django_tables2",
    "maintenance_mode",
    "reversion",
    "phonenumber_field",
    "django_prometheus",
    "django_select2",
    "templated_email",
    "html2text",
    "django_otp.plugins.otp_totp",
    "django_otp.plugins.otp_static",
    "django_otp.plugins.otp_email",
    "django_otp",
    "aleksis.core",
    "allauth",
    "allauth.account",
    "allauth.socialaccount",
    "invitations",
    "health_check",
    "health_check.db",
    "health_check.cache",
    "health_check.storage",
    "health_check.contrib.psutil",
    "health_check.contrib.migrations",
    "dynamic_preferences",
    "dynamic_preferences.users.apps.UserPreferencesConfig",
    "impersonate",
    "two_factor",
    "two_factor.plugins.email",
    "two_factor.plugins.phonenumber",
    "two_factor.plugins.webauthn",
    "material",
    "ckeditor",
    "ckeditor_uploader",
    "colorfield",
    "django_bleach",
    "favicon",
    "django_filters",
    "oauth2_provider",
    "rest_framework",
    "graphene_django",
    "dj_iconify.apps.DjIconifyConfig",
    "recurrence",
]

merge_app_settings("INSTALLED_APPS", INSTALLED_APPS, True)
INSTALLED_APPS += get_app_packages()

STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "sass_processor.finders.CssFinder",
]

MIDDLEWARE = [
    #    'django.middleware.cache.UpdateCacheMiddleware',
    "django_prometheus.middleware.PrometheusBeforeMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.http.ConditionalGetMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django_otp.middleware.OTPMiddleware",
    "impersonate.middleware.ImpersonateMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "maintenance_mode.middleware.MaintenanceModeMiddleware",
    "aleksis.core.util.middlewares.EnsurePersonMiddleware",
    "django_prometheus.middleware.PrometheusAfterMiddleware",
    "allauth.account.middleware.AccountMiddleware",
    #    'django.middleware.cache.FetchFromCacheMiddleware'
]

ROOT_URLCONF = "aleksis.core.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "maintenance_mode.context_processors.maintenance_mode",
                "dynamic_preferences.processors.global_preferences",
                "aleksis.core.util.core_helpers.custom_information_processor",
                "aleksis.core.util.context_processors.need_maintenance_response_context_processor",
            ],
        },
    },
]

# Attention: The following context processors must accept None
# as first argument (in addition to a HttpRequest object)
NON_REQUEST_CONTEXT_PROCESSORS = [
    "django.template.context_processors.i18n",
    "django.template.context_processors.tz",
    "aleksis.core.util.core_helpers.custom_information_processor",
]

WSGI_APPLICATION = "aleksis.core.wsgi.application"

# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django_prometheus.db.backends.postgresql",
        "NAME": _settings.get("database.name", "aleksis"),
        "USER": _settings.get("database.username", "aleksis"),
        "PASSWORD": _settings.get("database.password", None),
        "HOST": _settings.get("database.host", "127.0.0.1"),
        "PORT": _settings.get("database.port", "5432"),
        "CONN_MAX_AGE": _settings.get("database.conn_max_age", None),
        "CONN_HEALTH_CHECK": True,
        "OPTIONS": _settings.get("database.options", {}),
    }
}

# Duplicate default database for out-of-transaction updates
DATABASES["default_oot"] = DATABASES["default"].copy()
DATABASE_ROUTERS = [
    "aleksis.core.util.core_helpers.OOTRouter",
]
DATABASE_OOT_LABELS = ["django_celery_results"]

merge_app_settings("DATABASES", DATABASES, False)

PASSWORD_HASHERS = [
    "django.contrib.auth.hashers.ScryptPasswordHasher",
    "django.contrib.auth.hashers.PBKDF2PasswordHasher",
    "django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher",
]

REDIS_HOST = _settings.get("redis.host", "localhost")
REDIS_PORT = _settings.get("redis.port", 6379)
REDIS_DB = _settings.get("redis.database", 0)
REDIS_PASSWORD = _settings.get("redis.password", None)
REDIS_USER = _settings.get("redis.user", None if REDIS_PASSWORD is None else "default")

REDIS_URL = (
    f"redis://{REDIS_USER+':'+REDIS_PASSWORD+'@' if REDIS_USER else ''}"
    f"{REDIS_HOST}:{REDIS_PORT}/{REDIS_DB}"
)

if _settings.get("caching.redis.enabled", not IN_PYTEST):
    CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.redis.RedisCache",
            "LOCATION": _settings.get("caching.redis.address", REDIS_URL),
        }
    }
else:
    CACHES = {
        "default": {
            # Use uWSGI if available (will auot-fallback to LocMemCache)
            "BACKEND": "django_uwsgi.cache.UwsgiCache"
        }
    }

INSTALLED_APPS.append("cachalot")
CACHALOT_TIMEOUT = _settings.get("caching.cachalot.timeout", None)
CACHALOT_DATABASES = set(["default", "default_oot"])
SILENCED_SYSTEM_CHECKS += ["cachalot.W001"]
CACHALOT_ENABLED = _settings.get("caching.query_caching", True)
CACHALOT_UNCACHABLE_TABLES = ("django_migrations", "django_session")

SESSION_ENGINE = "django.contrib.sessions.backends.cached_db"
SESSION_CACHE_ALIAS = "default"

# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

AUTH_INITIAL_SUPERUSER = {
    "username": _settings.get("auth.superuser.username", "admin"),
    "password": _settings.get("auth.superuser.password", "admin"),
    "email": _settings.get("auth.superuser.email", "root@example.com"),
}

# Authentication backends are dynamically populated
AUTHENTICATION_BACKENDS = []

# Configuration for django-allauth.

# Use custom adapter to override some behaviour, i.e. honour the LDAP backend
SOCIALACCOUNT_ADAPTER = "aleksis.core.util.auth_helpers.OurSocialAccountAdapter"

# Get django-allauth providers from config
_SOCIALACCOUNT_PROVIDERS = _settings.get("auth.providers", None)
if _SOCIALACCOUNT_PROVIDERS:
    SOCIALACCOUNT_PROVIDERS = _SOCIALACCOUNT_PROVIDERS.to_dict()

    # Add configured social auth providers to INSTALLED_APPS
    for provider, config in SOCIALACCOUNT_PROVIDERS.items():
        INSTALLED_APPS.append(f"allauth.socialaccount.providers.{provider}")
        SOCIALACCOUNT_PROVIDERS[provider] = {k.upper(): v for k, v in config.items()}


# Configure custom forms

ACCOUNT_FORMS = {
    "signup": "aleksis.core.forms.AccountRegisterForm",
}

# Use custom adapter
ACCOUNT_ADAPTER = "aleksis.core.util.auth_helpers.OurAccountAdapter"

# Require password confirmation
SIGNUP_PASSWORD_ENTER_TWICE = True

# Allow login by either username or email
ACCOUNT_AUTHENTICATION_METHOD = _settings.get("auth.registration.method", "username_email")

# Require email address to sign up
ACCOUNT_EMAIL_REQUIRED = _settings.get("auth.registration.email_required", True)
SOCIALACCOUNT_EMAIL_REQUIRED = False

# Cooldown for verification mails
ACCOUNT_EMAIL_CONFIRMATION_COOLDOWN = _settings.get("auth.registration.verification_cooldown", 180)

# Require email verification after sign up
ACCOUNT_EMAIL_VERIFICATION = _settings.get("auth.registration.email_verification", "optional")
SOCIALACCOUNT_EMAIL_VERIFICATION = False

# Email subject prefix for verification mails
ACCOUNT_EMAIL_SUBJECT_PREFIX = _settings.get("auth.registration.subject", "[AlekSIS] ")

# Max attempts before login timeout
ACCOUNT_LOGIN_ATTEMPTS_LIMIT = _settings.get("auth.login.login_limit", 5)

# Login timeout after max attempts in seconds
ACCOUNT_LOGIN_ATTEMPTS_TIMEOUT = _settings.get("auth.login.login_timeout", 300)

# Email confirmation field in form
ACCOUNT_SIGNUP_EMAIL_ENTER_TWICE = True

# Enforce uniqueness of email addresses
ACCOUNT_UNIQUE_EMAIL = _settings.get("auth.login.registration.unique_email", True)

# Configurable username validators
ACCOUNT_USERNAME_VALIDATORS = "aleksis.core.util.auth_helpers.custom_username_validators"

# Configuration for django-invitations

# Use custom account adapter
ACCOUNT_ADAPTER = "invitations.models.InvitationsAdapter"
# Expire invitations are configured amout of days
INVITATIONS_INVITATION_EXPIRY = _settings.get("auth.invitation.expiry", 3)
# Use email prefix configured for django-allauth
INVITATIONS_EMAIL_SUBJECT_PREFIX = ACCOUNT_EMAIL_SUBJECT_PREFIX
# Use custom invitation model
INVITATIONS_INVITATION_MODEL = "core.PersonInvitation"
# Use custom invitation form
INVITATIONS_INVITE_FORM = "aleksis.core.forms.PersonCreateInviteForm"
# Display error message if invitation code is invalid
INVITATIONS_GONE_ON_ACCEPT_ERROR = False
# Mark invitation as accepted after signup
INVITATIONS_ACCEPT_INVITE_AFTER_SIGNUP = True

# Configuration for OAuth2 provider
OAUTH2_PROVIDER = {
    "SCOPES_BACKEND_CLASS": "aleksis.core.util.auth_helpers.AppScopes",
    "OAUTH2_VALIDATOR_CLASS": "aleksis.core.util.auth_helpers.CustomOAuth2Validator",
    "OIDC_ENABLED": True,
    "OIDC_ISS_ENDPOINT": BASE_URL,
    "REFRESH_TOKEN_EXPIRE_SECONDS": _settings.get("oauth2.token_expiry", 86400),
    "PKCE_REQUIRED": False,
}
OAUTH2_PROVIDER_APPLICATION_MODEL = "core.OAuthApplication"
OAUTH2_PROVIDER_GRANT_MODEL = "core.OAuthGrant"
OAUTH2_PROVIDER_ACCESS_TOKEN_MODEL = "core.OAuthAccessToken"  # noqa: S105
OAUTH2_PROVIDER_ID_TOKEN_MODEL = "core.OAuthIDToken"  # noqa: S105
OAUTH2_PROVIDER_REFRESH_TOKEN_MODEL = "core.OAuthRefreshToken"  # noqa: S105

_OIDC_RSA_KEY_DEFAULT = "/etc/aleksis/oidc.pem"
_OIDC_RSA_KEY = _settings.get("oauth2.oidc.rsa_key", "/etc/aleksis/oidc.pem")
if "BEGIN RSA PRIVATE KEY" in _OIDC_RSA_KEY:
    OAUTH2_PROVIDER["OIDC_RSA_PRIVATE_KEY"] = _OIDC_RSA_KEY
elif _OIDC_RSA_KEY == _OIDC_RSA_KEY_DEFAULT and not os.path.exists(_OIDC_RSA_KEY):
    warnings.warn(
        (
            f"The default OIDC RSA key in {_OIDC_RSA_KEY} does not exist. "
            f"RSA will be disabled for now, but creating and configuring a "
            f"key is recommended. To silence this warning, set oauth2.oidc.rsa_key "
            f"to the empty string in a configuration file."
        )
    )
elif _OIDC_RSA_KEY:
    with open(_OIDC_RSA_KEY, "r") as f:
        OAUTH2_PROVIDER["OIDC_RSA_PRIVATE_KEY"] = f.read()

# Configuration for REST framework
REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "oauth2_provider.contrib.rest_framework.OAuth2Authentication",
    ]
}

# Configuration for GraphQL framework
GRAPHENE = {
    "SCHEMA": "aleksis.core.schema.schema",
}

# LDAP config
if _settings.get("ldap.uri", None):
    # LDAP dependencies are not necessarily installed, so import them here
    import ldap  # noqa
    from django_auth_ldap.config import (
        LDAPSearch,
        LDAPSearchUnion,
        NestedGroupOfNamesType,
        NestedGroupOfUniqueNamesType,
        PosixGroupType,
    )

    AUTH_LDAP_GLOBAL_OPTIONS = {
        ldap.OPT_NETWORK_TIMEOUT: _settings.get("ldap.network_timeout", 3),
    }

    # Enable Django's integration to LDAP
    AUTHENTICATION_BACKENDS.append("aleksis.core.util.ldap.LDAPBackend")

    AUTH_LDAP_SERVER_URI = _settings.get("ldap.uri")

    # Optional: non-anonymous bind
    if _settings.get("ldap.bind.dn", None):
        AUTH_LDAP_BIND_DN = _settings.get("ldap.bind.dn")
        AUTH_LDAP_BIND_PASSWORD = _settings.get("ldap.bind.password")

    # Keep local password for users to be required to provide their old password on change
    AUTH_LDAP_SET_USABLE_PASSWORD = _settings.get("ldap.handle_passwords", True)

    # Keep bound as the authenticating user
    # Ensures proper read permissions, and ability to change password without admin
    AUTH_LDAP_BIND_AS_AUTHENTICATING_USER = True

    # The TOML config might contain either one table or an array of tables
    _AUTH_LDAP_USER_SETTINGS = _settings.get("ldap.users.search")
    if not isinstance(_AUTH_LDAP_USER_SETTINGS, list):
        _AUTH_LDAP_USER_SETTINGS = [_AUTH_LDAP_USER_SETTINGS]

    # Search attributes to find users by username
    AUTH_LDAP_USER_SEARCH = LDAPSearchUnion(
        *[
            LDAPSearch(
                entry["base"],
                ldap.SCOPE_SUBTREE,
                entry.get("filter", "(uid=%(user)s)"),
            )
            for entry in _AUTH_LDAP_USER_SETTINGS
        ]
    )

    # Mapping of LDAP attributes to Django model fields
    AUTH_LDAP_USER_ATTR_MAP = {
        "first_name": _settings.get("ldap.users.map.first_name", "givenName"),
        "last_name": _settings.get("ldap.users.map.last_name", "sn"),
        "email": _settings.get("ldap.users.map.email", "mail"),
    }

    # Discover flags by LDAP groups
    if _settings.get("ldap.groups.search", None):
        group_type = _settings.get("ldap.groups.type", "groupOfNames")

        # The TOML config might contain either one table or an array of tables
        _AUTH_LDAP_GROUP_SETTINGS = _settings.get("ldap.groups.search")
        if not isinstance(_AUTH_LDAP_GROUP_SETTINGS, list):
            _AUTH_LDAP_GROUP_SETTINGS = [_AUTH_LDAP_GROUP_SETTINGS]

        AUTH_LDAP_GROUP_SEARCH = LDAPSearchUnion(
            *[
                LDAPSearch(
                    entry["base"],
                    ldap.SCOPE_SUBTREE,
                    entry.get("filter", f"(objectClass={group_type})"),
                )
                for entry in _AUTH_LDAP_GROUP_SETTINGS
            ]
        )

        _group_type = _settings.get("ldap.groups.type", "groupOfNames").lower()
        if _group_type == "groupofnames":
            AUTH_LDAP_GROUP_TYPE = NestedGroupOfNamesType()
        elif _group_type == "groupofuniquenames":
            AUTH_LDAP_GROUP_TYPE = NestedGroupOfUniqueNamesType()
        elif _group_type == "posixgroup":
            AUTH_LDAP_GROUP_TYPE = PosixGroupType()

        AUTH_LDAP_USER_FLAGS_BY_GROUP = {}
        for _flag in ["is_active", "is_staff", "is_superuser"]:
            _dn = _settings.get(f"ldap.groups.flags.{_flag}", None)
            if _dn:
                AUTH_LDAP_USER_FLAGS_BY_GROUP[_flag] = _dn

        # Backend admin requires superusers to also be staff members
        if (
            "is_superuser" in AUTH_LDAP_USER_FLAGS_BY_GROUP
            and "is_staff" not in AUTH_LDAP_USER_FLAGS_BY_GROUP
        ):
            AUTH_LDAP_USER_FLAGS_BY_GROUP["is_staff"] = AUTH_LDAP_USER_FLAGS_BY_GROUP[
                "is_superuser"
            ]

# Add ModelBackend last so all other backends get a chance
# to verify passwords first
AUTHENTICATION_BACKENDS.append("django.contrib.auth.backends.ModelBackend")

# Authentication backend for django-allauth.
AUTHENTICATION_BACKENDS.append("allauth.account.auth_backends.AuthenticationBackend")

# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGES = [
    ("en", _("English")),
    ("de", _("German")),
    ("uk", _("Ukrainian")),
]
LANGUAGE_CODE = _settings.get("l10n.lang", "en")
TIME_ZONE = _settings.get("l10n.tz", "UTC")
USE_TZ = True

PHONENUMBER_DEFAULT_REGION = _settings.get("l10n.phone_number_country", None)

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/


STATIC_URL = _settings.get("static.url", "/static/")
MEDIA_URL = _settings.get("media.url", "/media/")

LOGIN_REDIRECT_URL = "index"
LOGOUT_REDIRECT_URL = "index"

STATIC_ROOT = _settings.get("static.root", os.path.join(BASE_DIR, "static"))
MEDIA_ROOT = _settings.get("media.root", os.path.join(BASE_DIR, "media"))
NODE_MODULES_ROOT = CACHE_DIR

YARN_INSTALLED_APPS = [
    "cleave.js@^1.6.0",
    "@fontsource/roboto@^4.5.5",
    "jquery@^3.6.0",
    "@materializecss/materialize@~1.0.0",
    "material-design-icons-iconfont@^6.7.0",
    "select2-materialize@^0.1.8",
    "paper-css@^0.4.1",
    "jquery-sortablejs@^1.0.1",
    "sortablejs@^1.15.0",
    "@sentry/tracing@^7.28.0",
    "luxon@^2.3.2",
    "@iconify/iconify@^2.2.1",
    "@iconify/json@^2.1.30",
    "@mdi/font@^7.2.96",
    "apollo-boost@^0.4.9",
    "apollo-link-batch-http@^1.2.14",
    "apollo-link-retry@^2.2.16",
    "apollo3-cache-persist@^0.14.1",
    "deepmerge@^4.2.2",
    "graphql@^15.8.0",
    "graphql-tag@^2.12.6",
    "sass@^1.32",
    "vue@^2.7.7",
    "vue-apollo@^3.1.0",
    "vuetify@^2.6.7",
    "vue-router@^3.5.2",
    "vue-cookies@^1.8.2",
    "vite@^4.0.1",
    "vite-plugin-pwa@^0.14.1",
    "vite-plugin-top-level-await@^1.2.2",
    "@vitejs/plugin-vue2@^2.2.0",
    "@rollup/plugin-node-resolve@^15.0.1",
    "@rollup/plugin-graphql@^2.0.2",
    "@rollup/plugin-virtual@^3.0.1",
    "rollup-plugin-license@^3.0.1",
    "vue-i18n@^8.0.0",
    "browserslist-to-esbuild@^1.2.0",
    "vite-plugin-browserslist-useragent@^0.4.1",
    "@sentry/vue@^7.28.0",
    "vue-draggable-grid@^0.4.0",
    "rrule",
    "luxon@^3.4.3",
]

merge_app_settings("YARN_INSTALLED_APPS", YARN_INSTALLED_APPS, True)

JS_URL = _settings.get("js_assets.url", STATIC_URL)
JS_ROOT = _settings.get("js_assets.root", os.path.join(NODE_MODULES_ROOT, "node_modules"))

DJANGO_VITE_ASSETS_PATH = os.path.join(NODE_MODULES_ROOT, "vite_bundles")
DJANGO_VITE_DEV_MODE = DEBUG

STATICFILES_DIRS = (
    DJANGO_VITE_ASSETS_PATH,
    JS_ROOT,
)

ANY_JS = {
    "materialize": {"js_url": JS_URL + "/@materializecss/materialize/dist/js/materialize.min.js"},
    "jQuery": {"js_url": JS_URL + "/jquery/dist/jquery.min.js"},
    "material-design-icons": {
        "css_url": JS_URL + "/material-design-icons-iconfont/dist/material-design-icons.css"
    },
    "paper-css": {"css_url": JS_URL + "/paper-css/paper.min.css"},
    "select2-materialize": {
        "css_url": JS_URL + "/select2-materialize/select2-materialize.css",
        "js_url": JS_URL + "/select2-materialize/index.js",
    },
    "sortablejs": {"js_url": JS_URL + "/sortablejs/Sortable.min.js"},
    "jquery-sortablejs": {"js_url": JS_URL + "/jquery-sortablejs/jquery-sortable.js"},
    "Roboto100": {"css_url": JS_URL + "/@fontsource/roboto/100.css"},
    "Roboto300": {"css_url": JS_URL + "/@fontsource/roboto/300.css"},
    "Roboto400": {"css_url": JS_URL + "/@fontsource/roboto/400.css"},
    "Roboto500": {"css_url": JS_URL + "/@fontsource/roboto/500.css"},
    "Roboto700": {"css_url": JS_URL + "/@fontsource/roboto/700.css"},
    "Roboto900": {"css_url": JS_URL + "/@fontsource/roboto/900.css"},
    "Sentry": {"js_url": JS_URL + "/@sentry/tracing/build/bundle.tracing.js"},
    "cleavejs": {"js_url": JS_URL + "/cleave.js/dist/cleave.min.js"},
    "luxon": {"js_url": JS_URL + "/luxon/build/global/luxon.min.js"},
    "iconify": {"js_url": JS_URL + "/@iconify/iconify/dist/iconify.min.js"},
}

merge_app_settings("ANY_JS", ANY_JS, True)

CLEAVE_JS = ANY_JS["cleavejs"]["js_url"]

SASS_PROCESSOR_ENABLED = True
SASS_PROCESSOR_AUTO_INCLUDE = False
SASS_PROCESSOR_CUSTOM_FUNCTIONS = {
    "get-colour": "aleksis.core.util.sass_helpers.get_colour",
    "get-preference": "aleksis.core.util.sass_helpers.get_preference",
}
SASS_PROCESSOR_INCLUDE_DIRS = [
    _settings.get(
        "materialize.sass_path", os.path.join(JS_ROOT, "@materializecss", "materialize", "sass")
    ),
    os.path.join(STATIC_ROOT, "public"),
]

ICONIFY_JSON_ROOT = os.path.join(JS_ROOT, "@iconify", "json")
ICONIFY_COLLECTIONS_ALLOWED = ["mdi"]

ADMINS = _settings.get(
    "contact.admins", [(AUTH_INITIAL_SUPERUSER["username"], AUTH_INITIAL_SUPERUSER["email"])]
)
SERVER_EMAIL = _settings.get("contact.from", ADMINS[0][1])
DEFAULT_FROM_EMAIL = _settings.get("contact.from", ADMINS[0][1])
MANAGERS = _settings.get("contact.admins", ADMINS)

if _settings.get("mail.server.host", None):
    EMAIL_HOST = _settings.get("mail.server.host")
    EMAIL_USE_TLS = _settings.get("mail.server.tls", False)
    EMAIL_USE_SSL = _settings.get("mail.server.ssl", False)
    if _settings.get("mail.server.port", None):
        EMAIL_PORT = _settings.get("mail.server.port")
    if _settings.get("mail.server.user", None):
        EMAIL_HOST_USER = _settings.get("mail.server.user")
        EMAIL_HOST_PASSWORD = _settings.get("mail.server.password")

TEMPLATED_EMAIL_BACKEND = "templated_email.backends.vanilla_django"
TEMPLATED_EMAIL_AUTO_PLAIN = True

DYNAMIC_PREFERENCES = {
    "REGISTRY_MODULE": "preferences",
}

MAINTENANCE_MODE = _settings.get("maintenance.enabled", None)
MAINTENANCE_MODE_IGNORE_IP_ADDRESSES = _settings.get(
    "maintenance.ignore_ips", _settings.get("maintenance.internal_ips", [])
)
MAINTENANCE_MODE_GET_CLIENT_IP_ADDRESS = "aleksis.core.util.core_helpers.get_ip"
MAINTENANCE_MODE_IGNORE_SUPERUSER = True
MAINTENANCE_MODE_STATE_FILE_NAME = _settings.get(
    "maintenance.statefile", "maintenance_mode_state.txt"
)
MAINTENANCE_MODE_STATE_BACKEND = "maintenance_mode.backends.DefaultStorageBackend"

DBBACKUP_STORAGE = _settings.get("backup.storage", "django.core.files.storage.FileSystemStorage")
DBBACKUP_STORAGE_OPTIONS = {"location": _settings.get("backup.location", "/var/backups/aleksis")}
DBBACKUP_CLEANUP_KEEP = _settings.get("backup.database.keep", 10)
DBBACKUP_CLEANUP_KEEP_MEDIA = _settings.get("backup.media.keep", 10)
DBBACKUP_GPG_RECIPIENT = _settings.get("backup.gpg_recipient", None)
DBBACKUP_COMPRESS_DB = _settings.get("backup.database.compress", True)
DBBACKUP_ENCRYPT_DB = _settings.get("backup.database.encrypt", DBBACKUP_GPG_RECIPIENT is not None)
DBBACKUP_COMPRESS_MEDIA = _settings.get("backup.media.compress", True)
DBBACKUP_ENCRYPT_MEDIA = _settings.get("backup.media.encrypt", DBBACKUP_GPG_RECIPIENT is not None)
DBBACKUP_CLEANUP_DB = _settings.get("backup.database.clean", True)
DBBACKUP_CLEANUP_MEDIA = _settings.get("backup.media.clean", True)
DBBACKUP_CONNECTOR_MAPPING = {
    "django_prometheus.db.backends.postgresql": "dbbackup.db.postgresql.PgDumpConnector",
}

if _settings.get("backup.storage.type", "").lower() == "s3":
    DBBACKUP_STORAGE = "storages.backends.s3.S3Storage"

    DBBACKUP_STORAGE_OPTIONS = {
        key: value for (key, value) in _settings.get("backup.storage.s3").items()
    }

IMPERSONATE = {"REQUIRE_SUPERUSER": True, "ALLOW_SUPERUSER": True, "REDIRECT_FIELD_NAME": "next"}

DJANGO_TABLES2_TEMPLATE = "django_tables2/materialize.html"

ANONYMIZE_ENABLED = _settings.get("maintenance.anonymisable", True)

LOGIN_URL = "two_factor:login"

if _settings.get("2fa.call.enabled", False):
    if "two_factor.middleware.threadlocals.ThreadLocals" not in MIDDLEWARE:
        MIDDLEWARE.insert(
            MIDDLEWARE.index("django_otp.middleware.OTPMiddleware") + 1,
            "two_factor.middleware.threadlocals.ThreadLocals",
        )
    TWO_FACTOR_CALL_GATEWAY = "two_factor.gateways.twilio.gateway.Twilio"

if _settings.get("2fa.sms.enabled", False):
    if "two_factor.middleware.threadlocals.ThreadLocals" not in MIDDLEWARE:
        MIDDLEWARE.insert(
            MIDDLEWARE.index("django_otp.middleware.OTPMiddleware") + 1,
            "two_factor.middleware.threadlocals.ThreadLocals",
        )
    TWO_FACTOR_SMS_GATEWAY = "two_factor.gateways.twilio.gateway.Twilio"

if _settings.get("twilio.sid", None):
    TWILIO_ACCOUNT_SID = _settings.get("twilio.sid")
    TWILIO_AUTH_TOKEN = _settings.get("twilio.token")
    TWILIO_CALLER_ID = _settings.get("twilio.callerid")

TWO_FACTOR_WEBAUTHN_RP_NAME = _settings.get("2fa.webauthn.rp_name", "AlekSIS")

CELERY_BROKER_URL = _settings.get("celery.broker", REDIS_URL)
CELERY_RESULT_BACKEND = "django-db"
CELERY_CACHE_BACKEND = "django-cache"
CELERY_BEAT_SCHEDULER = "django_celery_beat.schedulers:DatabaseScheduler"
CELERY_RESULT_EXTENDED = True

if _settings.get("celery.email", False):
    EMAIL_BACKEND = "djcelery_email.backends.CeleryEmailBackend"

if _settings.get("dev.uwsgi.celery", DEBUG):
    concurrency = _settings.get("celery.uwsgi.concurrency", 2)
    UWSGI.setdefault("attach-daemon", [])
    UWSGI["attach-daemon"].append(f"celery -A aleksis.core worker --concurrency={concurrency}")
    UWSGI["attach-daemon"].append("celery -A aleksis.core beat")
    UWSGI["attach-daemon"].append("aleksis-admin vite --no-install serve")

DEFAULT_FAVICON_PATHS = {
    "pwa_icon": os.path.join(STATIC_ROOT, "img/aleksis-icon-maskable.png"),
    "favicon": os.path.join(STATIC_ROOT, "img/aleksis-favicon.png"),
}
PWA_ICONS_CONFIG = {
    "android": [192, 512],
    "apple": [76, 114, 152, 180],
    "apple_splash": [192],
    "microsoft": [144],
}
FAVICON_PATH = os.path.join("public", "favicon")
FAVICON_CONFIG = {
    "shortcut icon": [16, 32, 48, 128, 192],
    "touch-icon": [196],
    "icon": [196],
}

SERVICE_WORKER_PATH = os.path.join(STATIC_ROOT, "sw.js")

CKEDITOR_CONFIGS = {
    "default": {
        "toolbar_Basic": [["Source", "-", "Bold", "Italic"]],
        "toolbar_Full": [
            {
                "name": "document",
                "items": ["Source", "-", "Save", "NewPage", "Preview", "Print", "-", "Templates"],
            },
            {
                "name": "clipboard",
                "items": [
                    "Cut",
                    "Copy",
                    "Paste",
                    "PasteText",
                    "PasteFromWord",
                    "-",
                    "Undo",
                    "Redo",
                ],
            },
            {"name": "editing", "items": ["Find", "Replace", "-", "SelectAll"]},
            {
                "name": "insert",
                "items": [
                    "Image",
                    "Table",
                    "HorizontalRule",
                    "Smiley",
                    "SpecialChar",
                    "PageBreak",
                    "Iframe",
                ],
            },
            "/",
            {
                "name": "basicstyles",
                "items": [
                    "Bold",
                    "Italic",
                    "Underline",
                    "Strike",
                    "Subscript",
                    "Superscript",
                    "-",
                    "RemoveFormat",
                ],
            },
            {
                "name": "paragraph",
                "items": [
                    "NumberedList",
                    "BulletedList",
                    "-",
                    "Outdent",
                    "Indent",
                    "-",
                    "Blockquote",
                    "CreateDiv",
                    "-",
                    "JustifyLeft",
                    "JustifyCenter",
                    "JustifyRight",
                    "JustifyBlock",
                    "-",
                    "BidiLtr",
                    "BidiRtl",
                    "Language",
                ],
            },
            {"name": "links", "items": ["Link", "Unlink", "Anchor"]},
            "/",
            {"name": "styles", "items": ["Styles", "Format", "Font", "FontSize"]},
            {"name": "colors", "items": ["TextColor", "BGColor"]},
            {"name": "tools", "items": ["Maximize", "ShowBlocks"]},
            {"name": "about", "items": ["About"]},
            {
                "name": "customtools",
                "items": [
                    "Preview",
                    "Maximize",
                ],
            },
        ],
        "toolbar": "Full",
        "tabSpaces": 4,
        "extraPlugins": ",".join(
            [
                "uploadimage",
                "div",
                "autolink",
                "autoembed",
                "embedsemantic",
                "autogrow",
                # 'devtools',
                "widget",
                "lineutils",
                "clipboard",
                "dialog",
                "dialogui",
                "elementspath",
            ]
        ),
    }
}

# Upload path for CKEditor. Relative to MEDIA_ROOT.
CKEDITOR_UPLOAD_PATH = "ckeditor_uploads/"

# Which HTML tags are allowed
BLEACH_ALLOWED_TAGS = ["p", "b", "i", "u", "em", "strong", "a", "div"]

# Which HTML attributes are allowed
BLEACH_ALLOWED_ATTRIBUTES = ["href", "title", "style"]

# Which CSS properties are allowed in 'style' attributes (assuming
# style is an allowed attribute)
BLEACH_ALLOWED_STYLES = ["font-family", "font-weight", "text-decoration", "font-variant"]

# Strip unknown tags if True, replace with HTML escaped characters if
# False
BLEACH_STRIP_TAGS = True

# Strip comments, or leave them in.
BLEACH_STRIP_COMMENTS = True

LOGGING = deepcopy(DEFAULT_LOGGING)
# Set root logging level as default
LOGGING["root"] = {
    "handlers": ["console"],
    "level": _settings.get("logging.level", "WARNING"),
}
# Configure global log Format
LOGGING["formatters"]["verbose"] = {
    "format": "{asctime} {levelname} {name}[{process}]: {message}",
    "style": "{",
}
# Add null handler for selective silencing
LOGGING["handlers"]["null"] = {"class": "logging.NullHandler"}
# Make console logging independent of DEBUG
LOGGING["handlers"]["console"]["filters"].remove("require_debug_true")
# Use root log level for console
del LOGGING["handlers"]["console"]["level"]
# Use verbose log format for console
LOGGING["handlers"]["console"]["formatter"] = "verbose"
# Disable exception mails if not desired
if not _settings.get("logging.mail_admins", True):
    LOGGING["loggers"]["django"]["handlers"].remove("mail_admins")
# Disable mails on disaalowed host by default
if not _settings.get("logging.disallowed_host", False):
    LOGGING["loggers"]["django.security.DisallowedHost"] = {
        "handlers": ["null"],
        "propagate": False,
    }
# Configure logging explicitly for Celery
LOGGING["loggers"]["celery"] = {
    "handlers": ["console"],
    "level": _settings.get("logging.level", "WARNING"),
    "propagate": False,
}
# Set Django log levels
LOGGING["loggers"]["django"]["level"] = _settings.get("logging.level", "WARNING")
LOGGING["loggers"]["django.server"]["level"] = _settings.get("logging.level", "WARNING")

# Rules and permissions

GUARDIAN_RAISE_403 = True
ANONYMOUS_USER_NAME = None

SILENCED_SYSTEM_CHECKS.append("guardian.W001")

# Append authentication backends
AUTHENTICATION_BACKENDS.append("rules.permissions.ObjectPermissionBackend")

HAYSTACK_CONNECTIONS = {
    "default": {
        "ENGINE": "haystack_redis.RedisEngine",
        "PATH": REDIS_URL,
    },
}

HAYSTACK_SIGNAL_PROCESSOR = "celery_haystack.signals.CelerySignalProcessor"
CELERY_HAYSTACK_IGNORE_RESULT = True

HAYSTACK_SEARCH_RESULTS_PER_PAGE = 10

DJANGO_EASY_AUDIT_WATCH_REQUEST_EVENTS = False

HEALTH_CHECK = {
    "DISK_USAGE_MAX": _settings.get("health.disk_usage_max_percent", 90),
    "MEMORY_MIN": _settings.get("health.memory_min_mb", 500),
}

DBBACKUP_CHECK_SECONDS = _settings.get("backup.database.check_seconds", 7200)
MEDIABACKUP_CHECK_SECONDS = _settings.get("backup.media.check_seconds", 7200)

PROMETHEUS_EXPORT_MIGRATIONS = False
PROMETHEUS_METRICS_EXPORT_PORT = _settings.get("prometheus.metrics.port", None)
PROMETHEUS_METRICS_EXPORT_ADDRESS = _settings.get("prometheus.metrucs.address", None)

SECURE_PROXY_SSL_HEADER = ("REQUEST_SCHEME", "https")

FILE_UPLOAD_HANDLERS = [
    "django.core.files.uploadhandler.MemoryFileUploadHandler",
    "django.core.files.uploadhandler.TemporaryFileUploadHandler",
]

if _settings.get("storage.type", "").lower() == "s3":
    INSTALLED_APPS.append("storages")

    DEFAULT_FILE_STORAGE = "storages.backends.s3.S3Storage"
    FILE_UPLOAD_HANDLERS.remove("django.core.files.uploadhandler.MemoryFileUploadHandler")

    if _settings.get("storage.s3.static.enabled", False):
        STATICFILES_STORAGE = "storages.backends.s3.S3StaticStorage"
        AWS_STORAGE_BUCKET_NAME_STATIC = _settings.get("storage.s3.static.bucket_name", "")
        AWS_S3_MAX_AGE_SECONDS_CACHED_STATIC = _settings.get(
            "storage.s3.static.max_age_seconds", 24 * 60 * 60
        )

    AWS_REGION = _settings.get("storage.s3.region_name", "")
    AWS_ACCESS_KEY_ID = _settings.get("storage.s3.access_key", "")
    AWS_SECRET_ACCESS_KEY = _settings.get("storage.s3.secret_key", "")
    AWS_SESSION_TOKEN = _settings.get("storage.s3.session_token", "")
    AWS_STORAGE_BUCKET_NAME = _settings.get("storage.s3.bucket_name", "")
    AWS_LOCATION = _settings.get("storage.s3.location", "")
    AWS_S3_ADDRESSING_STYLE = _settings.get("storage.s3.addressing_style", "auto")
    AWS_S3_ENDPOINT_URL = _settings.get("storage.s3.endpoint_url", "")
    AWS_S3_KEY_PREFIX = _settings.get("storage.s3.key_prefix", "")
    AWS_S3_BUCKET_AUTH = _settings.get("storage.s3.bucket_auth", True)
    AWS_S3_MAX_AGE_SECONDS = _settings.get("storage.s3.max_age_seconds", 24 * 60 * 60)
    AWS_S3_PUBLIC_URL = _settings.get("storage.s3.public_url", "")
    AWS_S3_REDUCED_REDUNDANCY = _settings.get("storage.s3.reduced_redundancy", False)
    AWS_S3_CONTENT_DISPOSITION = _settings.get("storage.s3.content_disposition", "")
    AWS_S3_CONTENT_LANGUAGE = _settings.get("storage.s3.content_language", "")
    AWS_S3_METADATA = _settings.get("storage.s3.metadata", {})
    AWS_S3_ENCRYPT_KEY = _settings.get("storage.s3.encrypt_key", False)
    AWS_S3_KMS_ENCRYPTION_KEY_ID = _settings.get("storage.s3.kms_encryption_key_id", "")
    AWS_S3_GZIP = _settings.get("storage.s3.gzip", True)
    AWS_S3_SIGNATURE_VERSION = _settings.get("storage.s3.signature_version", None)
    AWS_S3_FILE_OVERWRITE = _settings.get("storage.s3.file_overwrite", False)
    AWS_S3_VERIFY = _settings.get("storage.s3.verify", True)
    AWS_S3_USE_SSL = _settings.get("storage.s3.use_ssl", True)
else:
    DEFAULT_FILE_STORAGE = "titofisto.TitofistoStorage"
    TITOFISTO_TIMEOUT = 10 * 60

TITOFISTO_ENABLE_UPLOAD = True
TITOFISTO_UPLOAD_NAMESPACE = "__titofisto__/upload/"

SASS_PROCESSOR_STORAGE = DEFAULT_FILE_STORAGE

SENTRY_ENABLED = _settings.get("health.sentry.enabled", False)
if SENTRY_ENABLED:
    import sentry_sdk
    from sentry_sdk.integrations.celery import CeleryIntegration
    from sentry_sdk.integrations.django import DjangoIntegration
    from sentry_sdk.integrations.redis import RedisIntegration

    from aleksis.core import __version__

    SENTRY_SETTINGS = {
        "dsn": _settings.get("health.sentry.dsn"),
        "environment": _settings.get("health.sentry.environment"),
        "traces_sample_rate": _settings.get("health.sentry.traces_sample_rate", 1.0),
        "send_default_pii": _settings.get("health.sentry.send_default_pii", False),
        "release": f"aleksis-core@{__version__}",
        "in_app_include": "aleksis",
    }
    sentry_sdk.init(
        integrations=[
            DjangoIntegration(transaction_style="function_name"),
            RedisIntegration(),
            CeleryIntegration(),
        ],
        **SENTRY_SETTINGS,
    )

SHELL_PLUS_MODEL_IMPORTS_RESOLVER = "django_extensions.collision_resolvers.AppLabelPrefixCR"
SHELL_PLUS_APP_PREFIXES = {
    "auth": "auth",
}
SHELL_PLUS_DONT_LOAD = []
merge_app_settings("SHELL_PLUS_APP_PREFIXES", SHELL_PLUS_APP_PREFIXES)
merge_app_settings("SHELL_PLUS_DONT_LOAD", SHELL_PLUS_DONT_LOAD)

X_FRAME_OPTIONS = "SAMEORIGIN"

# Add django-cleanup after all apps to ensure that it gets all signals as last app
INSTALLED_APPS.append("django_cleanup.apps.CleanupConfig")

locals().update(get_app_settings_overrides())

ABSOLUTE_URL_OVERRIDES = {
    "auth.user": lambda o: f"/admin/auth/user/{o.pk}/change",
}



import json
from datetime import datetime, time

from django.core.exceptions import PermissionDenied
from django.urls import reverse

import graphene
from graphene import ObjectType

from aleksis.core.mixins import CalendarEventMixin
from aleksis.core.util.core_helpers import has_person


class CalendarEventType(ObjectType):
    name = graphene.String()
    description = graphene.String()
    location = graphene.String(required=False)
    start = graphene.DateTime()
    end = graphene.DateTime()
    color = graphene.String()
    uid = graphene.String()
    all_day = graphene.Boolean()
    status = graphene.String()
    meta = graphene.String()

    def resolve_name(root, info, **kwargs):
        return root["SUMMARY"]

    def resolve_description(root, info, **kwargs):
        return root["DESCRIPTION"]

    def resolve_location(root, info, **kwargs):
        return root.get("LOCATION", "")

    def resolve_start(root, info, **kwargs):
        return root["DTSTART"].dt

    def resolve_end(root, info, **kwargs):
        return root["DTEND"].dt

    def resolve_color(root, info, **kwargs):
        return root["COLOR"]

    def resolve_uid(root, info, **kwargs):
        return root["UID"]

    def resolve_all_day(root, info, **kwargs):
        return not isinstance(root["DTSTART"].dt, datetime)

    def resolve_status(root, info, **kwargs):
        return root.get("STATUS", "")

    def resolve_meta(root, info, **kwargs):
        return root.get("X-META", "{}")


class CalendarType(ObjectType):
    name = graphene.String(required=True)
    verbose_name = graphene.String(required=True)
    description = graphene.String()
    events = graphene.List(
        CalendarEventType,
        start=graphene.Date(required=False),
        end=graphene.Date(required=False),
        params=graphene.String(required=False),
    )

    def resolve_events(root, info, start=None, end=None, params=None, **kwargs):
        if params:
            params = json.loads(params)
        feed = root.create_feed(info.context, params)
        if end:
            end = datetime.combine(end, time.max)

        return feed.get_single_events(start, end)

    color = graphene.String()

    url = graphene.String()

    activated = graphene.Boolean()

    def resolve_verbose_name(root, info, **kwargs):
        return root.get_verbose_name(info.context)

    def resolve_description(root, info, **kwargs):
        return root.get_description(info.context)

    def resolve_url(root, info, **kwargs):
        return info.context.build_absolute_uri(reverse("calendar_feed", args=[root.name]))

    def resolve_color(root, info, **kwargs):
        return root.get_color(info.context)

    def resolve_activated(root, info, **kwargs):
        return root.get_activated(info.context.user.person)


class SetCalendarStatusMutation(graphene.Mutation):
    """Mutation to change the status of a calendar."""

    class Arguments:
        calendars = graphene.List(graphene.String)

    ok = graphene.Boolean()

    def mutate(root, info, calendars, **kwargs):
        if not has_person(info.context):
            raise PermissionDenied
        calendar_feeds = [cal for cal in calendars if cal in CalendarEventMixin.valid_feed_names]
        info.context.user.person.preferences["calendar__activated_calendars"] = calendar_feeds
        return SetCalendarStatusMutation(ok=True)


class CalendarBaseType(ObjectType):
    calendar_feeds = graphene.List(
        CalendarType, names=graphene.List(graphene.String, required=False)
    )

    all_feeds_url = graphene.String()

    def resolve_calendar_feeds(root, info, names=None, **kwargs):
        if names is not None:
            feeds = [CalendarEventMixin.get_object_by_name(name) for name in names]
            return [feed for feed in feeds if feed.get_enabled(info.context)]
        return CalendarEventMixin.get_enabled_feeds(info.context)

    def resolve_all_feeds_url(root, info, **kwargs):
        return info.context.build_absolute_uri(reverse("all_calendar_feeds"))



from django.template.loader_tags import register


@register.inclusion_tag("components/msgbox.html")
def msg_box(msg, status="success", icon="info"):
    return {"msg": msg, "status": status, "icon": icon}



from collections import OrderedDict

from material import Layout, Row


class PreferenceLayout(Layout):
    """django-material Layout object for managing preferences."""

    def __init__(self, form_base_class, section=None):
        """
        Create Layout object for the given form_base_class.

        :param form_base_class: A Form class used as the base. Must have a ``registry` attribute
        :param section: A section where the layout builder will load preferences
        """
        registry = form_base_class.registry
        if section:
            # Try to use section param
            preferences_obj = registry.preferences(section=section)
        else:
            # display all preferences in the form
            preferences_obj = registry.preferences()

        rows = OrderedDict()

        for preference in preferences_obj:
            row_name = preference.get("row", preference.identifier())
            rows.setdefault(row_name, [])
            rows[row_name].append(preference.identifier())

        rows_material = []
        for fields in rows.values():
            rows_material.append(Row(*fields))
        super().__init__(*rows_material)



from django.conf import settings
from django.forms import EmailField, ImageField, URLField
from django.forms.widgets import SelectMultiple
from django.utils.translation import gettext_lazy as _

from colorfield.widgets import ColorWidget
from dynamic_preferences.preferences import Section
from dynamic_preferences.types import (
    BooleanPreference,
    ChoicePreference,
    FilePreference,
    IntegerPreference,
    LongStringPreference,
    ModelMultipleChoicePreference,
    MultipleChoicePreference,
    StringPreference,
)
from oauth2_provider.models import AbstractApplication

from .mixins import CalendarEventMixin, PublicFilePreferenceMixin
from .models import Group, Person
from .registries import person_preferences_registry, site_preferences_registry
from .util.notifications import get_notification_choices_lazy

general = Section("general", verbose_name=_("General"))
school = Section("school", verbose_name=_("School"))
theme = Section("theme", verbose_name=_("Theme"))
mail = Section("mail", verbose_name=_("Mail"))
notification = Section("notification", verbose_name=_("Notifications"))
footer = Section("footer", verbose_name=_("Footer"))
account = Section("account", verbose_name=_("Accounts"))
auth = Section("auth", verbose_name=_("Authentication"))
internationalisation = Section("internationalisation", verbose_name=_("Internationalisation"))
calendar = Section("calendar", verbose_name=_("Calendar"))


@site_preferences_registry.register
class SiteTitle(StringPreference):
    """Title of the AlekSIS instance, e.g. schools display name."""

    section = general
    name = "title"
    default = "AlekSIS"
    verbose_name = _("Site title")
    required = True


@site_preferences_registry.register
class SiteDescription(StringPreference):
    """Site description, e.g. a slogan."""

    section = general
    name = "description"
    default = "The Free School Information System"
    required = False
    verbose_name = _("Site description")


@site_preferences_registry.register
class ColourPrimary(StringPreference):
    """Primary colour in AlekSIS frontend."""

    section = theme
    name = "primary"
    default = "#0d5eaf"
    verbose_name = _("Primary colour")
    widget = ColorWidget
    required = True


@site_preferences_registry.register
class ColourSecondary(StringPreference):
    """Secondary colour in AlekSIS frontend."""

    section = theme
    name = "secondary"
    default = "#0d5eaf"
    verbose_name = _("Secondary colour")
    widget = ColorWidget
    required = True


@site_preferences_registry.register
class Logo(PublicFilePreferenceMixin, FilePreference):
    """Logo of your AlekSIS instance."""

    section = theme
    field_class = ImageField
    name = "logo"
    verbose_name = _("Logo")
    required = False


@site_preferences_registry.register
class Favicon(PublicFilePreferenceMixin, FilePreference):
    """Favicon of your AlekSIS instance."""

    section = theme
    field_class = ImageField
    name = "favicon"
    verbose_name = _("Favicon")
    required = False


@site_preferences_registry.register
class PWAIcon(PublicFilePreferenceMixin, FilePreference):
    """PWA-Icon of your AlekSIS instance."""

    section = theme
    field_class = ImageField
    name = "pwa_icon"
    verbose_name = _("PWA-Icon")
    required = False


@site_preferences_registry.register
class PWAIconMaskable(BooleanPreference):
    """PWA icon is maskable."""

    section = theme
    name = "pwa_icon_maskable"
    verbose_name = _("PWA-Icon is maskable")
    default = True
    required = False


@site_preferences_registry.register
class MailOutName(StringPreference):
    """Mail out name of your AlekSIS instance."""

    section = mail
    name = "name"
    default = "AlekSIS"
    verbose_name = _("Mail out name")
    required = True


@site_preferences_registry.register
class MailOut(StringPreference):
    """Mail out address of your AlekSIS instance."""

    section = mail
    name = "address"
    default = settings.DEFAULT_FROM_EMAIL
    verbose_name = _("Mail out address")
    field_class = EmailField
    required = True


@site_preferences_registry.register
class PrivacyURL(StringPreference):
    """Link to privacy policy of your AlekSIS instance."""

    section = footer
    name = "privacy_url"
    default = ""
    required = False
    verbose_name = _("Link to privacy policy")
    field_class = URLField


@site_preferences_registry.register
class ImprintURL(StringPreference):
    """Link to imprint of your AlekSIS instance."""

    section = footer
    name = "imprint_url"
    default = ""
    required = False
    verbose_name = _("Link to imprint")
    field_class = URLField


@person_preferences_registry.register
class AdressingNameFormat(ChoicePreference):
    """User preference for adressing name format."""

    section = notification
    name = "addressing_name_format"
    default = "first_last"
    verbose_name = _("Name format for addressing")
    choices = (
        ("first_last", "John Doe"),
        ("last_fist", "Doe, John"),
    )
    required = True


@person_preferences_registry.register
class NotificationChannels(ChoicePreference):
    """User preference for notification channels."""

    # FIXME should be a MultipleChoicePreference
    section = notification
    name = "channels"
    default = "email"
    required = False
    verbose_name = _("Channels to use for notifications")
    choices = get_notification_choices_lazy()


@person_preferences_registry.register
class Design(ChoicePreference):
    """Change design (on supported pages)."""

    section = theme
    name = "design"
    default = "light"
    verbose_name = _("Select Design")
    choices = [
        # ("system", _("System Design")),
        ("light", _("Light mode")),
        # ("dark", _("Dark mode")),
    ]


@site_preferences_registry.register
class PrimaryGroupPattern(StringPreference):
    """Regular expression to match primary group."""

    section = account
    name = "primary_group_pattern"
    default = ""
    required = False
    verbose_name = _("Regular expression to match primary group, e.g. '^Class .*'")


@site_preferences_registry.register
class PrimaryGroupField(ChoicePreference):
    """Field on person to match primary group against."""

    section = account
    name = "primary_group_field"
    default = "name"
    required = False
    verbose_name = _("Field on person to match primary group against")

    def get_choices(self):
        return Person.syncable_fields_choices()


@site_preferences_registry.register
class AutoCreatePerson(BooleanPreference):
    section = account
    name = "auto_create_person"
    default = False
    required = False
    verbose_name = _("Automatically create new persons for new users")


@site_preferences_registry.register
class AutoLinkPerson(BooleanPreference):
    section = account
    name = "auto_link_person"
    default = False
    required = False
    verbose_name = _("Automatically link existing persons to new users by their e-mail address")


@site_preferences_registry.register
class SchoolName(StringPreference):
    """Display name of the school."""

    section = school
    name = "name"
    default = ""
    required = False
    verbose_name = _("Display name of the school")


@site_preferences_registry.register
class SchoolNameOfficial(StringPreference):
    """Official name of the school, e.g. as given by supervisory authority."""

    section = school
    name = "name_official"
    default = ""
    required = False
    verbose_name = _("Official name of the school, e.g. as given by supervisory authority")


@site_preferences_registry.register
class AllowPasswordChange(BooleanPreference):
    section = auth
    name = "allow_password_change"
    default = True
    verbose_name = _("Allow users to change their passwords")


@site_preferences_registry.register
class AllowPasswordReset(BooleanPreference):
    section = auth
    name = "allow_password_reset"
    default = True
    verbose_name = _("Allow users to reset their passwords")


@site_preferences_registry.register
class SignupEnabled(BooleanPreference):
    section = auth
    name = "signup_enabled"
    default = False
    verbose_name = _("Enable signup")


@site_preferences_registry.register
class AllowedUsernameRegex(StringPreference):
    section = auth
    name = "allowed_username_regex"
    default = ".+"
    verbose_name = _("Regular expression for allowed usernames")


@site_preferences_registry.register
class InviteEnabled(BooleanPreference):
    section = auth
    name = "invite_enabled"
    default = False
    verbose_name = _("Enable invitations")


@site_preferences_registry.register
class InviteCodeLength(IntegerPreference):
    section = auth
    name = "invite_code_length"
    default = 3
    verbose_name = _("Length of invite code. (Default 3: abcde-acbde-abcde)")


@site_preferences_registry.register
class InviteCodePacketSize(IntegerPreference):
    section = auth
    name = "invite_code_packet_size"
    default = 5
    verbose_name = _("Size of packets. (Default 5: abcde)")


@site_preferences_registry.register
class OAuthAllowedGrants(MultipleChoicePreference):
    """Grant Flows allowed for OAuth applications."""

    section = auth
    name = "oauth_allowed_grants"
    default = [grant[0] for grant in AbstractApplication.GRANT_TYPES]
    widget = SelectMultiple
    verbose_name = _("Allowed Grant Flows for OAuth applications")
    field_attribute = {"initial": []}
    choices = AbstractApplication.GRANT_TYPES
    required = False


@site_preferences_registry.register
class DataChecksSendEmails(BooleanPreference):
    """Enable email sending if data checks detect problems."""

    section = general
    name = "data_checks_send_emails"
    default = False
    verbose_name = _("Send emails if data checks detect problems")


@site_preferences_registry.register
class DataChecksEmailsRecipients(ModelMultipleChoicePreference):
    """Email recipients for data check problem emails."""

    section = general
    name = "data_checks_recipients"
    default = []
    model = Person
    verbose_name = _("Email recipients for data checks problem emails")


@site_preferences_registry.register
class DataChecksEmailsRecipientGroups(ModelMultipleChoicePreference):
    """Email recipient groups for data check problem emails."""

    section = general
    name = "data_checks_recipient_groups"
    default = []
    model = Group
    verbose_name = _("Email recipient groups for data checks problem emails")


@site_preferences_registry.register
class AnonymousDashboard(BooleanPreference):
    section = general
    name = "anonymous_dashboard"
    default = False
    required = False
    verbose_name = _("Show dashboard to users without login")


@site_preferences_registry.register
class DashboardEditing(BooleanPreference):
    section = general
    name = "dashboard_editing"
    default = True
    required = False
    verbose_name = _("Allow users to edit their dashboard")


@site_preferences_registry.register
class EditableFieldsPerson(MultipleChoicePreference):
    """Fields on person model that should be editable by the person."""

    section = account
    name = "editable_fields_person"
    default = []
    widget = SelectMultiple
    verbose_name = _("Fields on person model which are editable by themselves.")
    field_attribute = {"initial": []}
    choices = [(field.name, field.name) for field in Person.syncable_fields()]
    required = False


@site_preferences_registry.register
class SendNotificationOnPersonChange(MultipleChoicePreference):
    """Fields on the person model that should trigger a notification on change."""

    section = account
    name = "notification_on_person_change"
    default = []
    widget = SelectMultiple
    verbose_name = _(
        "Editable fields on person model which should trigger a notification on change"
    )
    field_attribute = {"initial": []}
    choices = [(field.name, field.name) for field in Person.syncable_fields()]
    required = False


@site_preferences_registry.register
class PersonChangeNotificationContact(StringPreference):
    """Mail recipient address for change notifications."""

    section = account
    name = "person_change_notification_contact"
    default = ""
    verbose_name = _("Contact for notification if a person changes their data")
    required = False


@site_preferences_registry.register
class PersonPreferPhoto(BooleanPreference):
    """Preference, whether personal photos should be displayed instead of avatars."""

    section = account
    name = "person_prefer_photo"
    default = False
    verbose_name = _("Prefer personal photos over avatars")


@site_preferences_registry.register
class PDFFileExpirationDuration(IntegerPreference):
    """PDF file expiration duration."""

    section = general
    name = "pdf_expiration"
    default = 3
    verbose_name = _("PDF file expiration duration")
    help_text = _("in minutes")


@person_preferences_registry.register
class AutoUpdatingDashboard(BooleanPreference):
    """User preference for automatically updating the dashboard."""

    section = general
    name = "automatically_update_dashboard"
    default = True
    verbose_name = _("Automatically update the dashboard and its widgets")


@site_preferences_registry.register
class AutoUpdatingDashboardSite(BooleanPreference):
    """Automatic updating of dashboard."""

    section = general
    name = "automatically_update_dashboard_site"
    default = True
    verbose_name = _("Automatically update the dashboard and its widgets sitewide")


@site_preferences_registry.register
class BirthdayFeedColor(StringPreference):
    """Color for the birthdays calendar feed."""

    section = calendar
    name = "birthday_color"
    default = "#0d5eaf"
    verbose_name = _("Birthday calendar feed color")
    widget = ColorWidget
    required = True


@site_preferences_registry.register
class HolidayFeedColor(StringPreference):
    """Color for the holidays calendar feed."""

    section = calendar
    name = "holiday_color"
    default = "#0d5eaf"
    verbose_name = _("Holiday calendar feed color")
    widget = ColorWidget
    required = True


@site_preferences_registry.register
class PersonalEventFeedColor(StringPreference):
    """Color for the personal events calendar feed."""

    section = calendar
    name = "personal_event_color"
    default = "#0d5eaf"
    verbose_name = _("Personal events feed color")
    widget = ColorWidget
    required = True


@person_preferences_registry.register
class ActivatedCalendars(MultipleChoicePreference):
    """Calendars that are activated for a person."""

    section = calendar
    name = "activated_calendars"
    default = []
    widget = SelectMultiple
    verbose_name = _("Activated calendars")
    required = False

    field_attribute = {"initial": []}
    choices = [(feed.name, feed.verbose_name) for feed in CalendarEventMixin.valid_feeds]


@site_preferences_registry.register
class DisallowedUids(LongStringPreference):
    section = auth
    name = "disallowed_uids"
    default = (
        "bin,daemon,Debian-exim,freerad,games,gnats,irc,list,lp,mail,man,messagebus,news,"
        "nslcd,ntp,openldap,postfix,postgres,proxy,root,sshd,sssd,statd,sync,sys,systemd-bus-proxy,"
        "systemd-network,systemd-resolve,systemd-timesync,uucp,www-data,"
        "webmaster,hostmaster,postmaster"
    )
    required = False
    verbose_name = _("Comma-separated list of disallowed usernames")



# Generated by Django 4.2.7 on 2023-11-08 13:00

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0055_customevent"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="CustomEvent",
            new_name="PersonalEvent",
        ),
    ]



import json
from typing import Any, Optional, Union

from django import template
from django.contrib.contenttypes.models import ContentType
from django.db.models import Model

register = template.Library()


@register.filter
def get_dict(value: Any, arg: Any) -> Any:
    """Get an attribute of an object dynamically from a string name."""
    if hasattr(value, str(arg)):
        return getattr(value, arg)
    elif hasattr(value, "keys") and arg in value:
        return value[arg]
    elif str(arg).isnumeric() and len(value) > int(arg):
        return value[int(arg)]
    else:
        return None


@register.simple_tag
def verbose_name(app_label: str, model: str, field: Optional[str] = None) -> str:
    """Get a verbose name of a model or a field by app label and model name."""
    ct = ContentType.objects.get(app_label=app_label, model=model).model_class()

    if field:
        # Field
        return ct._meta.get_field(field).verbose_name.title()
    else:
        # Whole model
        return ct._meta.verbose_name.title()


@register.simple_tag
def verbose_name_object(model: Model, field: Optional[str] = None) -> str:
    """Get a verbose name of a model or a field by a model or an instance of a model."""
    if field:
        # Field
        return model._meta.get_field(field).verbose_name.title()
    else:
        # Whole model
        return model._meta.verbose_name.title()


@register.simple_tag
def parse_json(value: Optional[str] = None) -> Union[dict, None]:
    """Template tag for parsing JSON from a string."""
    if not value:
        return None
    return json.loads(value)



# Generated by Django 4.2.8 on 2023-12-10 12:44

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0060_person_unique_short_name_email'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='group',
            name='additional_fields',
        ),
        migrations.DeleteModel(
            name='AdditionalField',
        ),
    ]



# Generated by Django 3.2.15 on 2022-11-20 14:20

from django.apps import apps
import aleksis.core.managers
from django.db import migrations, models
from django.db.utils import ProgrammingError
import django.db.models.deletion
import django.utils.timezone
import oauth2_provider.generators
import oauth2_provider.models
from psycopg2.errors import UndefinedTable


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0046_notification_create_field_icon'),
    ]

    # This migration must run after Chronos' migration 1 through 12, but before
    #  13. That's because we are in fact moving a model, and we need to make sure
    #  that this migration runs at the right time.
    if "chronos" in apps.app_configs:
        recorder = migrations.recorder
        applied = False
        try:
            applied = recorder.MigrationRecorder.Migration.objects.filter(app="core", name="0047_add_room_model").exists()
        except ProgrammingError:
            applied = False
        if not applied:
            dependencies.append(('chronos', '0012_add_supervision_global_permission'))

    operations = [
        migrations.CreateModel(
            name='Room',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('short_name', models.CharField(max_length=255, verbose_name='Short name')),
                ('name', models.CharField(max_length=255, verbose_name='Long name')),
            ],
            options={
                'verbose_name': 'Room',
                'verbose_name_plural': 'Rooms',
                'ordering': ['name', 'short_name'],
                'permissions': (('view_room_timetable', 'Can view room timetable'),),
            },
            managers=[
                ('objects', aleksis.core.managers.AlekSISBaseManager()),
            ],
        ),
        migrations.AddConstraint(
            model_name='room',
            constraint=models.UniqueConstraint(fields=('short_name',), name='unique_room_short_name_per_site'),
        ),
        # Migrate data from Chronos table; deletion will be handled by Chronos
        migrations.RunSQL(
            """
            -- Copy rooms from chronos if table exists
            DO $$BEGIN INSERT INTO core_room SELECT * FROM chronos_room; EXCEPTION WHEN undefined_table THEN NULL; END$$;
            """
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-19 23:10

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0054_create_organisation_model"),
    ]

    operations = [
        migrations.CreateModel(
            name="CustomEvent",
            fields=[
                (
                    "calendarevent_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.calendarevent",
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                ("description", models.TextField(blank=True, verbose_name="Description")),
                ("location", models.CharField(blank=True, max_length=255, verbose_name="Location")),
                ("groups", models.ManyToManyField(blank=True, related_name="+", to="core.group")),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="owned_events",
                        to="core.person",
                        verbose_name="Owner",
                    ),
                ),
                ("persons", models.ManyToManyField(blank=True, related_name="+", to="core.person")),
            ],
            options={
                "abstract": False,
            },
            bases=("core.calendarevent",),
        ),
    ]



import json
import os
from datetime import datetime, timedelta
from importlib import import_module, metadata
from itertools import groupby
from operator import itemgetter
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Sequence, Union
from warnings import warn

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files import File
from django.db.models import Model, QuerySet
from django.http import HttpRequest
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.functional import lazy
from django.utils.module_loading import import_string

import django_ical.feedgenerator as feedgenerator
import recurring_ical_events
from cachalot.api import invalidate
from cachalot.signals import post_invalidation
from cache_memoize import cache_memoize
from icalendar import Calendar, Event, Todo

if TYPE_CHECKING:
    from django.contrib.contenttypes.models import ContentType

    from favicon.models import Favicon


def copyright_years(years: Sequence[int], separator: str = ", ", joiner: str = "–") -> str:
    """Take a sequence of integers and produces a string with ranges.

    >>> copyright_years([1999, 2000, 2001, 2005, 2007, 2008, 2009])
    '1999–2001, 2005, 2007–2009'
    """
    ranges = [
        list(map(itemgetter(1), group))
        for _, group in groupby(enumerate(years), lambda e: e[1] - e[0])
    ]
    years_strs = [
        str(range_[0]) if len(range_) == 1 else joiner.join([str(range_[0]), str(range_[-1])])
        for range_ in ranges
    ]

    return separator.join(years_strs)


def get_app_packages(only_official: bool = False) -> Sequence[str]:
    """Find all registered apps from the setuptools entrypoint."""
    apps = []

    for ep in metadata.entry_points().get("aleksis.app", []):
        path = f"{ep.module}.{ep.attr}"
        if path.startswith("aleksis.apps.") or not only_official:
            apps.append(path)

    return apps


def get_app_module(app: str, name: str) -> Optional[ModuleType]:
    """Get a named module of an app."""
    pkg = ".".join(app.split(".")[:-2])

    while "." in pkg:
        try:
            return import_module(f"{pkg}.{name}")
        except ImportError:
            # Import errors are non-fatal.
            pkg = ".".join(pkg.split(".")[:-1])

    # The app does not have this module
    return None


def merge_app_settings(
    setting: str, original: Union[dict, list], deduplicate: bool = False
) -> Union[dict, list]:
    """Merge app settings.

    Get a named settings constant from all apps and merge it into the original.
    To use this, add a settings.py file to the app, in the same format as Django's
    main settings.py.

    Note: Only selected names will be imported frm it to minimise impact of
    potentially malicious apps!
    """
    for app in get_app_packages():
        mod_settings = get_app_module(app, "settings")
        if not mod_settings:
            continue

        app_setting = getattr(mod_settings, setting, None)
        if not app_setting:
            # The app might not have this setting or it might be empty. Ignore it in that case.
            continue

        for entry in app_setting:
            if entry in original:
                if not deduplicate:
                    raise AttributeError(f"{entry} already set in original.")
            else:
                if isinstance(original, list):
                    original.append(entry)
                elif isinstance(original, dict):
                    original[entry] = app_setting[entry]
                else:
                    raise TypeError("Only dict and list settings can be merged.")


def get_app_settings_overrides() -> dict[str, Any]:
    """Get app settings overrides.

    Official apps (those under the ``aleksis.apps` namespace) can override
    or add settings by listing them in their ``settings.overrides``.
    """
    overrides = {}

    for app in get_app_packages(True):
        mod_settings = get_app_module(app, "settings")
        if not mod_settings:
            continue

        if hasattr(mod_settings, "overrides"):
            for name in mod_settings.overrides:
                overrides[name] = getattr(mod_settings, name)

    return overrides


def get_site_preferences():
    """Get the preferences manager of the current site."""
    from ..registries import site_preferences_registry  # noqa

    return site_preferences_registry.manager()


def lazy_preference(section: str, name: str) -> Callable[[str, str], Any]:
    """Lazily get a config value from dynamic preferences.

    Useful to bind preferences
    to other global settings to make them available to third-party apps that are not
    aware of dynamic preferences.
    """

    def _get_preference(section: str, name: str) -> Any:
        return get_site_preferences()[f"{section}__{name}"]

    # The type is guessed from the default value to improve lazy()'s behaviour
    # FIXME Reintroduce the behaviour described above
    return lazy(_get_preference, str)(section, name)


def get_or_create_favicon(title: str, default: str, is_favicon: bool = False) -> "Favicon":
    """Ensure that there is always a favicon object."""
    from favicon.models import Favicon  # noqa

    if not os.path.exists(default):
        warn("staticfiles are not ready yet, not creating default icons")
        return
    elif os.path.isdir(default):
        raise ImproperlyConfigured(f"staticfiles are broken: unexpected directory at {default}")

    favicon, created = Favicon.on_site.get_or_create(
        title=title, defaults={"isFavicon": is_favicon}
    )

    changed = False

    if favicon.isFavicon != is_favicon:
        favicon.isFavicon = True
        changed = True

    if created:
        with open(default, "rb") as f:
            favicon.faviconImage.save(os.path.basename(default), File(f))
        changed = True

    if changed:
        favicon.save()

    return favicon


def get_pwa_icons():
    from django.conf import settings  # noqa

    favicon = get_or_create_favicon("pwa_icon", settings.DEFAULT_FAVICON_PATHS["pwa_icon"])
    favicon_imgs = favicon.get_favicons(config_override=settings.PWA_ICONS_CONFIG)
    return favicon_imgs


def is_impersonate(request: HttpRequest) -> bool:
    """Check whether the user was impersonated by an admin."""
    if hasattr(request, "user"):
        return getattr(request.user, "is_impersonate", False)
    else:
        return False


def has_person(obj: Union[HttpRequest, Model]) -> bool:
    """Check wehether a model object has a person attribute linking it to a Person object.

    The passed object can also be a HttpRequest object, in which case its
    associated User object is unwrapped and tested.
    """
    if isinstance(obj, HttpRequest):
        if hasattr(obj, "user"):
            obj = obj.user
        else:
            return False

    if obj.is_anonymous:
        return False

    person = getattr(obj, "person", None)
    if person is None or getattr(person, "is_dummy", False):
        return False

    return True


def custom_information_processor(request: Union[HttpRequest, None]) -> dict:
    """Provide custom information in all templates."""
    pwa_icons = get_pwa_icons()
    regrouped_pwa_icons = {}
    for pwa_icon in pwa_icons:
        regrouped_pwa_icons.setdefault(pwa_icon.rel, {})
        regrouped_pwa_icons[pwa_icon.rel][pwa_icon.size] = pwa_icon

    # This dictionary is passed to the frontend and made available as
    #  `$root.$aleksisFrontendSettings` in Vue.
    frontend_settings = {
        "sentry": {
            "enabled": settings.SENTRY_ENABLED,
        },
        "urls": {
            "base": settings.BASE_URL,
            "graphql": reverse("graphql"),
        },
    }

    context = {
        "ADMINS": settings.ADMINS,
        "PWA_ICONS": regrouped_pwa_icons,
        "SENTRY_ENABLED": settings.SENTRY_ENABLED,
        "SITE_PREFERENCES": get_site_preferences(),
        "BASE_URL": settings.BASE_URL,
        "FRONTEND_SETTINGS": frontend_settings,
    }

    if settings.SENTRY_ENABLED:
        frontend_settings["sentry"].update(settings.SENTRY_SETTINGS)

        import sentry_sdk

        span = sentry_sdk.Hub.current.scope.span
        if span is not None:
            context["SENTRY_TRACE_ID"] = span.to_traceparent()

    return context


def now_tomorrow() -> datetime:
    """Return current time tomorrow."""
    return timezone.now() + timedelta(days=1)


def objectgetter_optional(
    model: Model, default: Optional[Any] = None, default_eval: bool = False
) -> Callable[[HttpRequest, Optional[int]], Model]:
    """Get an object by pk, defaulting to None."""

    def get_object(request: HttpRequest, id_: Optional[int] = None, **kwargs) -> Optional[Model]:
        if id_ is not None:
            return get_object_or_404(model, pk=id_)
        else:
            try:
                return eval(default) if default_eval else default  # noqa:S307
            except (AttributeError, KeyError, IndexError):
                return None

    return get_object


@cache_memoize(3600)
def get_content_type_by_perm(perm: str) -> Union["ContentType", None]:
    from django.contrib.contenttypes.models import ContentType  # noqa

    try:
        return ContentType.objects.get(
            app_label=perm.split(".", 1)[0], permission__codename=perm.split(".", 1)[1]
        )
    except ContentType.DoesNotExist:
        return None


@cache_memoize(3600)
def queryset_rules_filter(
    obj: Union[HttpRequest, Model], queryset: QuerySet, perm: str
) -> QuerySet:
    """Filter queryset by user and permission."""
    wanted_objects = set()
    if isinstance(obj, HttpRequest) and hasattr(obj, "user"):
        obj = obj.user

    for item in queryset:
        if obj.has_perm(perm, item):
            wanted_objects.add(item.pk)

    return queryset.filter(pk__in=wanted_objects)


def generate_random_code(length, packet_size) -> str:
    """Generate random code for e.g. invitations."""
    return get_random_string(packet_size * length).lower()


def monkey_patch() -> None:  # noqa
    """Monkey-patch dependencies for special behaviour."""
    # Unwrap promises in JSON serializer instead of stringifying
    from django.core.serializers import json
    from django.utils.functional import Promise

    class DjangoJSONEncoder(json.DjangoJSONEncoder):
        def default(self, o: Any) -> Any:
            if isinstance(o, Promise) and hasattr(o, "copy"):
                return o.copy()
            return super().default(o)

    json.DjangoJSONEncoder = DjangoJSONEncoder


def get_allowed_object_ids(request: HttpRequest, models: list) -> list:
    """Get all objects of all given models the user of a given request is allowed to view."""
    allowed_object_ids = []

    for model in models:
        app_label = model._meta.app_label
        model_name = model.__name__.lower()

        # Loop through the pks of all objects of the current model the user is allowed to view
        # and put the corresponding ids into a django-haystack-style-formatted list
        allowed_object_ids += [
            f"{app_label}.{model_name}.{pk}"
            for pk in queryset_rules_filter(
                request, model.objects.all(), f"{app_label}.view_{model_name}_rule"
            ).values_list("pk", flat=True)
        ]

    return allowed_object_ids


def process_custom_context_processors(context_processors: list) -> Dict[str, Any]:
    """Process custom context processors."""
    context = {}
    processors = tuple(import_string(path) for path in context_processors)
    for processor in processors:
        context.update(processor(None))
    return context


def create_default_celery_schedule():
    """Create default periodic tasks in database for tasks that have a schedule defined."""
    from celery import current_app
    from celery.schedules import BaseSchedule, crontab, schedule, solar
    from django_celery_beat.clockedschedule import clocked
    from django_celery_beat.models import (
        ClockedSchedule,
        CrontabSchedule,
        IntervalSchedule,
        PeriodicTask,
        SolarSchedule,
    )

    defined_periodic_tasks = PeriodicTask.objects.values_list("task", flat=True).all()

    for name, task in current_app.tasks.items():
        if name in defined_periodic_tasks:
            # Task is already known in database, skip
            continue

        run_every = getattr(task, "run_every", None)
        if not run_every:
            # Task has no default schedule, skip
            continue

        if isinstance(run_every, (float, int, timedelta)):
            # Schedule is defined as a raw seconds value or timedelta, convert to schedule class
            run_every = schedule(run_every)
        elif not isinstance(run_every, BaseSchedule):
            raise ValueError(f"Task {name} has an invalid schedule defined.")

        # Find matching django-celery-beat schedule model
        if isinstance(run_every, clocked):
            Schedule = ClockedSchedule
            attr = "clocked"
        elif isinstance(run_every, crontab):
            Schedule = CrontabSchedule
            attr = "crontab"
        elif isinstance(run_every, schedule):
            Schedule = IntervalSchedule
            attr = "interval"
        elif isinstance(run_every, solar):
            Schedule = SolarSchedule
            attr = "solar"
        else:
            raise ValueError(f"Task {name} has an unknown schedule class defined.")

        # Get or create schedule in database
        db_schedule = Schedule.from_schedule(run_every)
        db_schedule.save()

        # Create periodic task
        PeriodicTask.objects.create(
            name=f"{name} (default schedule)", task=name, **{attr: db_schedule}
        )


class OOTRouter:
    """Database router for operations that should run out of transaction.

    This router routes database operations for certain apps through
    the separate default_oot connection, to ensure that data get
    updated immediately even during atomic transactions.
    """

    default_db = "default"
    oot_db = "default_oot"

    _cachalot_invalidating = []

    @property
    def oot_labels(self):
        return settings.DATABASE_OOT_LABELS

    @property
    def default_dbs(self):
        return set((self.default_db, self.oot_db))

    def is_same_db(self, db1: str, db2: str):
        return set((db1, db2)).issubset(self.default_dbs)

    def db_for_read(self, model: Model, **hints) -> Optional[str]:
        if model._meta.app_label in self.oot_labels:
            return self.oot_db

        return None

    def db_for_write(self, model: Model, **hints) -> Optional[str]:
        return self.db_for_read(model, **hints)

    def allow_relation(self, obj1: Model, obj2: Model, **hints) -> Optional[bool]:
        # Allow relations between default database and OOT connection
        # They are the same database
        if self.is_same_db(obj1._state.db, obj2._state.db):
            return True

        return None

    def allow_migrate(
        self, db: str, app_label: str, model_name: Optional[str] = None, **hints
    ) -> Optional[bool]:
        # Never allow any migrations on the default_oot database
        # It connects to the same database as default, so everything
        # migrated there
        if db == self.oot_db:
            return False

        return None

    @classmethod
    def _invalidate_cachalot(cls, sender, **kwargs):
        if sender in cls._cachalot_invalidating:
            return
        cls._cachalot_invalidating.append(sender)

        if kwargs["db_alias"] == cls.default_db:
            invalidate(sender, db_alias=cls.oot_db)
        elif kwargs["db_alias"] == cls.oot_db:
            invalidate(sender, db_alias=cls.default_db)

        if sender in cls._cachalot_invalidating:
            cls._cachalot_invalidating.remove(sender)


post_invalidation.connect(OOTRouter._invalidate_cachalot)


def get_ip(*args, **kwargs):
    """Recreate ipware.ip.get_ip as it was replaced by get_client_ip."""
    from ipware.ip import get_client_ip  # noqa

    return get_client_ip(*args, **kwargs)[0]


feedgenerator.FEED_FIELD_MAP = feedgenerator.FEED_FIELD_MAP + (("color", "color"),)
feedgenerator.ITEM_ELEMENT_FIELD_MAP = feedgenerator.ITEM_ELEMENT_FIELD_MAP + (
    ("color", "color"),
    ("meta", "x-meta"),
    ("reference_object", "reference_object"),
)


class ExtendedICal20Feed(feedgenerator.ICal20Feed):
    """Extends the ICal20Feed class from django-ical.

    Adds a method to return the actual calendar object.
    """

    def get_calendar_object(self, with_meta: bool = True, with_reference_object: bool = False):
        cal = Calendar()
        cal.add("version", "2.0")
        cal.add("calscale", "GREGORIAN")

        for ifield, efield in feedgenerator.FEED_FIELD_MAP:
            val = self.feed.get(ifield)
            if val is not None:
                cal.add(efield, val)

        self.write_items(cal, with_meta=with_meta, with_reference_object=with_reference_object)

        return cal

    def write(self, outfile, encoding):
        cal = self.get_calendar_object(with_meta=False)

        to_ical = getattr(cal, "as_string", None)
        if not to_ical:
            to_ical = cal.to_ical
        outfile.write(to_ical())

    def write_items(self, calendar, with_meta: bool = True, with_reference_object: bool = False):
        for item in self.items:
            component_type = item.get("component_type")
            element = Todo() if component_type == "todo" else Event()

            for ifield, efield in feedgenerator.ITEM_ELEMENT_FIELD_MAP:
                val = item.get(ifield)
                if val is not None:
                    if ifield == "attendee":
                        for list_item in val:
                            element.add(efield, list_item)
                    elif ifield == "valarm":
                        for list_item in val:
                            element.add_component(list_item)
                    elif ifield == "meta":
                        if with_meta:
                            element.add(efield, json.dumps(val))
                    elif ifield == "reference_object":
                        if with_reference_object:
                            element.add(efield, val, encode=False)
                    else:
                        element.add(efield, val)
            calendar.add_component(element)

    def get_single_events(self, start=None, end=None, with_reference_object: bool = False):
        """Get single event objects for this feed."""
        events = recurring_ical_events.of(
            self.get_calendar_object(with_reference_object=with_reference_object)
        )
        if start and end:
            return events.between(start, end)
        return events.all()



from graphene_django import DjangoObjectType
from graphene_django_cud.mutations import (
    DjangoBatchCreateMutation,
    DjangoBatchDeleteMutation,
    DjangoBatchPatchMutation,
)

from ..models import Room
from .base import (
    DjangoFilterMixin,
    PermissionBatchDeleteMixin,
    PermissionBatchPatchMixin,
    PermissionsTypeMixin,
)


class RoomType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = Room
        fields = ("id", "name", "short_name")
        filter_fields = {
            "id": ["exact", "lte", "gte"],
            "name": ["icontains"],
            "short_name": ["icontains"],
        }

    @classmethod
    def get_queryset(cls, queryset, info):
        return queryset  # FIXME filter this queryset based on permissions


class RoomBatchCreateMutation(PermissionBatchPatchMixin, DjangoBatchCreateMutation):
    class Meta:
        model = Room
        permissions = ("core.create_room",)
        only_fields = ("id", "name", "short_name")


class RoomBatchDeleteMutation(PermissionBatchDeleteMixin, DjangoBatchDeleteMutation):
    class Meta:
        model = Room
        permissions = ("core.delete_room",)


class RoomBatchPatchMutation(PermissionBatchPatchMixin, DjangoBatchPatchMutation):
    class Meta:
        model = Room
        permissions = ("core.change_room",)
        only_fields = ("id", "name", "short_name")



import graphene
from graphene import ObjectType


class AppURLType(ObjectType):
    name = graphene.String(required=True)
    url = graphene.String(required=True)


class CopyrightType(ObjectType):
    years = graphene.String(required=True)
    name = graphene.String(required=True)
    email = graphene.String(required=True)


class LicenceFlagsType(ObjectType):
    isFsfLibre = graphene.Boolean(required=True)
    isOsiApproved = graphene.Boolean(required=True)


class SubLicenceType(ObjectType):
    isDeprecatedLicenseId = graphene.Boolean(default_value=False)
    isFsfLibre = graphene.Boolean(default_value=False)
    isOsiApproved = graphene.Boolean(default_value=False)
    licenseId = graphene.String(required=True)
    name = graphene.String(required=True)
    referenceNumber = graphene.Int(default_value=-1)
    url = graphene.String()


class LicenceType(ObjectType):
    verbose_name = graphene.String(required=True)
    flags = graphene.Field(LicenceFlagsType, required=True)
    licences = graphene.List(SubLicenceType)


class AppType(ObjectType):
    copyrights = graphene.List(CopyrightType)
    licence = graphene.Field(LicenceType)
    name = graphene.String(required=True)
    verbose_name = graphene.String(required=True)
    version = graphene.String()
    urls = graphene.List(AppURLType)

    def resolve_verbose_name(root, info, **kwargs):
        return root.get_name()

    def resolve_version(root, info, **kwargs):
        return root.get_version()

    def resolve_licence(root, info, **kwargs):
        return root.get_licence_dict()

    def resolve_urls(root, info, **kwargs):
        return root.get_urls_dict()

    def resolve_copyrights(root, info, **kwargs):
        return root.get_copyright_dicts()



# flake8: noqa: DJ12

import os
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Iterable, List, Optional, Union

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.views import LoginView, RedirectURLMixin
from django.db import models
from django.db.models import JSONField
from django.db.models.fields import CharField, TextField
from django.forms.forms import BaseForm
from django.forms.models import ModelForm, ModelFormMetaclass, fields_for_model
from django.http import HttpRequest, HttpResponse
from django.utils.functional import classproperty, lazy
from django.utils.translation import gettext as _
from django.views.generic import CreateView, UpdateView
from django.views.generic.edit import DeleteView, ModelFormMixin

import reversion
from django_ical.feedgenerator import ITEM_ELEMENT_FIELD_MAP
from dynamic_preferences.settings import preferences_settings
from dynamic_preferences.types import FilePreference
from guardian.admin import GuardedModelAdmin
from guardian.core import ObjectPermissionChecker
from icalendar import Calendar
from material.base import Fieldset, Layout, LayoutNode
from polymorphic.base import PolymorphicModelBase
from polymorphic.models import PolymorphicModel
from rules.contrib.admin import ObjectPermissionsModelAdmin

from aleksis.core.managers import (
    AlekSISBaseManager,
    AlekSISBaseManagerWithoutMigrations,
    PolymorphicBaseManager,
    SchoolTermRelatedQuerySet,
)

from .util.core_helpers import ExtendedICal20Feed

if TYPE_CHECKING:
    from .models import Person


class _ExtensibleModelBase(models.base.ModelBase):
    """Ensure predefined behaviour on model creation.

    This metaclass serves the following purposes:

     - Register all AlekSIS models with django-reverseion
    """

    def __new__(mcls, name, bases, attrs):
        mcls = super().__new__(mcls, name, bases, attrs)

        if "Meta" not in attrs or not attrs["Meta"].abstract:
            # Register all non-abstract models with django-reversion
            mcls = reversion.register(mcls)

            mcls.extra_permissions = []

        return mcls


def _generate_one_to_one_proxy_property(field, subfield):
    def getter(self):
        if hasattr(self, field.name):
            related = getattr(self, field.name)
            return getattr(related, subfield.name)
        # Related instane does not exist
        return None

    def setter(self, val):
        if hasattr(self, field.name):
            related = getattr(self, field.name)
        else:
            # Auto-create related instance (but do not save)
            related = field.related_model()
            setattr(related, field.remote_field.name, self)
            # Ensure the related model is saved later
            self._save_reverse = getattr(self, "_save_reverse", []) + [related]
        setattr(related, subfield.name, val)

    return property(getter, setter)


class ExtensibleModel(models.Model, metaclass=_ExtensibleModelBase):
    """Base model for all objects in AlekSIS apps.

    This base model ensures all objects in AlekSIS apps fulfill the
    following properties:

     * `versions` property to retrieve all versions of the model from reversion
     * Allow injection of fields and code from AlekSIS apps to extend
       model functionality.

    Injection of fields and code
    ============================

    After all apps have been loaded, the code in the `model_extensions` module
    in every app is executed. All code that shall be injected into a model goes there.

    :Example:

    .. code-block:: python

       from datetime import date, timedelta

       from aleksis.core.models import Person

       @Person.property
       def is_cool(self) -> bool:
           return True

       @Person.property
       def age(self) -> timedelta:
           return self.date_of_birth - date.today()

    For a more advanced example, using features from the ORM, see AlekSIS-App-Chronos
    and AlekSIS-App-Alsijil.

    :Date: 2019-11-07
    :Authors:
        - Dominik George <dominik.george@teckids.org>
    """

    # Defines a material design icon associated with this type of model
    icon_ = "radiobox-blank"

    managed_by_app_label = models.CharField(
        max_length=255,
        verbose_name="App label of app responsible for managing this instance",
        editable=False,
        blank=True,
    )

    extended_data = JSONField(default=dict, editable=False)

    extra_permissions = []

    objects = AlekSISBaseManager()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """Ensure all functionality of our extensions that needs saving gets it."""
        # For auto-created remote syncable fields
        if hasattr(self, "_save_reverse"):
            for related in self._save_reverse:
                related.save()
            del self._save_reverse

        super().save(*args, **kwargs)

    def get_absolute_url(self) -> str:
        """Get the URL o a view representing this model instance."""
        pass

    @property
    def versions(self) -> list[tuple[str, tuple[Any, Any]]]:
        """Get all versions of this object from django-reversion.

        Includes diffs to previous version.
        """
        versions = reversion.models.Version.objects.get_for_object(self)

        versions_with_changes = []
        for i, version in enumerate(versions):
            diff = {}
            if i > 0:
                prev_version = versions[i - 1]

                for k, val in version.field_dict.items():
                    prev_val = prev_version.field_dict.get(k, None)
                    if prev_val != val:
                        diff[k] = (prev_val, val)

            versions_with_changes.append((version, diff))

        return versions_with_changes

    @classmethod
    def _safe_add(cls, obj: Any, name: Optional[str]) -> None:
        # Decide the name for the attribute
        if name is None:
            prop_name = obj.__name__
        else:
            if name.isidentifier():
                prop_name = name
            else:
                raise ValueError(f"{name} is not a valid name.")

        # Verify that attribute name does not clash with other names in the class
        if hasattr(cls, prop_name):
            raise ValueError(f"{prop_name} already used.")

        # Let Django's model magic add the attribute if we got here
        cls.add_to_class(name, obj)

    @classmethod
    def property_(cls, func: Callable[[], Any], name: Optional[str] = None) -> None:
        """Add the passed callable as a property."""
        cls._safe_add(property(func), name or func.__name__)

    @classmethod
    def method(cls, func: Callable[[], Any], name: Optional[str] = None) -> None:
        """Add the passed callable as a method."""
        cls._safe_add(func, name or func.__name__)

    @classmethod
    def class_method(cls, func: Callable[[], Any], name: Optional[str] = None) -> None:
        """Add the passed callable as a classmethod."""
        cls._safe_add(classmethod(func), name or func.__name__)

    @classmethod
    def get_filter_fields(cls) -> List[str]:
        """Get names of all text-searchable fields of this model."""
        fields = []
        for field in cls.syncable_fields():
            if isinstance(field, (CharField, TextField)):
                fields.append(field.name)
        return fields

    @classmethod
    def syncable_fields(
        cls, recursive: bool = True, exclude_remotes: list = None
    ) -> list[models.Field]:
        """Collect all fields that can be synced on a model.

        If recursive is True, it recurses into related models and generates virtual
        proxy fields to access fields in related models."""
        if not exclude_remotes:
            exclude_remotes = []

        fields = []
        for field in cls._meta.get_fields():
            if field.is_relation and field.one_to_one and recursive:
                if ExtensibleModel not in field.related_model.__mro__:
                    # Related model is not extensible and thus has no syncable fields
                    continue
                if field.related_model in exclude_remotes:
                    # Remote is excluded, probably to avoid recursion
                    continue

                # Recurse into related model to get its fields as well
                for subfield in field.related_model.syncable_fields(
                    recursive, exclude_remotes + [cls]
                ):
                    # generate virtual field names for proxy access
                    name = f"_{field.name}__{subfield.name}"
                    verbose_name = (
                        f"{field.name} ({field.related_model._meta.verbose_name})"
                        " → {subfield.verbose_name}"
                    )

                    if not hasattr(cls, name):
                        # Add proxy properties to handle access to related model
                        setattr(cls, name, _generate_one_to_one_proxy_property(field, subfield))

                    # Generate a fake field class with enough API to detect attribute names
                    fields.append(
                        type(
                            "FakeRelatedProxyField",
                            (),
                            {
                                "name": name,
                                "verbose_name": verbose_name,
                                "to_python": lambda v: subfield.to_python(v),  # noqa: B023
                            },
                        )
                    )
            elif field.editable and not field.auto_created:
                fields.append(field)

        return fields

    @classmethod
    def syncable_fields_choices(cls) -> tuple[tuple[str, str]]:
        """Collect all fields that can be synced on a model."""
        return tuple(
            [(field.name, field.verbose_name or field.name) for field in cls.syncable_fields()]
        )

    @classmethod
    def syncable_fields_choices_lazy(cls) -> Callable[[], tuple[tuple[str, str]]]:
        """Collect all fields that can be synced on a model."""
        return lazy(cls.syncable_fields_choices, tuple)

    @classmethod
    def add_permission(cls, name: str, verbose_name: str):
        """Dynamically add a new permission to a model."""
        cls.extra_permissions.append((name, verbose_name))

    def set_object_permission_checker(self, checker: ObjectPermissionChecker):
        """Annotate a ``ObjectPermissionChecker`` for use with permission system."""
        self._permission_checker = checker


class _ExtensiblePolymorphicModelBase(_ExtensibleModelBase, PolymorphicModelBase):
    """Base class for extensible, polymorphic models."""


class ExtensiblePolymorphicModel(
    ExtensibleModel, PolymorphicModel, metaclass=_ExtensiblePolymorphicModelBase
):
    """Model class for extensible, polymorphic models."""

    objects = PolymorphicBaseManager()

    class Meta:
        abstract = True


class PureDjangoModel:
    """No-op mixin to mark a model as deliberately not using ExtensibleModel."""

    pass


class GlobalPermissionModel(models.Model):
    """Base model for global permissions.

    This base model ensures that global permissions are not managed."""

    class Meta:
        default_permissions = ()
        abstract = True
        managed = False


class _ExtensibleFormMetaclass(ModelFormMetaclass):
    def __new__(cls, name, bases, dct):
        x = super().__new__(cls, name, bases, dct)

        # Enforce a default for the base layout for forms that o not specify one
        base_layout = x.layout.elements if hasattr(x, "layout") else []

        x.base_layout = base_layout
        x.layout = Layout(*base_layout)

        return x


class ExtensibleForm(ModelForm, metaclass=_ExtensibleFormMetaclass):
    """Base model for extensible forms.

    This mixin adds functionality which allows
    - apps to add layout nodes to the layout used by django-material

    :Add layout nodes:

    .. code-block:: python

        from material import Fieldset

        from aleksis.core.forms import ExampleForm

        node = Fieldset("field_name")
        ExampleForm.add_node_to_layout(node)

    """

    @classmethod
    def add_node_to_layout(cls, node: Union[LayoutNode, str]):
        """Add a node to `layout` attribute.

        :param node: django-material layout node (Fieldset, Row etc.)
        :type node: LayoutNode
        """
        cls.base_layout.append(node)
        cls.layout = Layout(*cls.base_layout)

        visit_nodes = [node]
        while visit_nodes:
            current_node = visit_nodes.pop()
            if isinstance(current_node, Fieldset):
                visit_nodes += node.elements
            else:
                field_name = (
                    current_node if isinstance(current_node, str) else current_node.field_name
                )
                field = fields_for_model(cls._meta.model, [field_name])[field_name]
                cls._meta.fields.append(field_name)
                cls.base_fields[field_name] = field
                setattr(cls, field_name, field)


class BaseModelAdmin(GuardedModelAdmin, ObjectPermissionsModelAdmin):
    """A base class for ModelAdmin combining django-guardian and rules."""

    pass


class SuccessMessageMixin(ModelFormMixin):
    success_message: Optional[str] = None

    def form_valid(self, form: BaseForm) -> HttpResponse:
        if self.success_message:
            messages.success(self.request, self.success_message)
        return super().form_valid(form)


class SuccessNextMixin(RedirectURLMixin):
    redirect_field_name = "next"

    def get_success_url(self) -> str:
        return LoginView.get_redirect_url(self) or super().get_success_url()


class AdvancedCreateView(SuccessMessageMixin, CreateView):
    pass


class AdvancedEditView(SuccessMessageMixin, UpdateView):
    pass


class AdvancedDeleteView(DeleteView):
    """Common confirm view for deleting.

    .. warning ::

        Using this view, objects are deleted permanently after confirming.
        We recommend to include the mixin :class:`reversion.views.RevisionMixin`
        from `django-reversion` to enable soft-delete.
    """

    success_message: Optional[str] = None

    def form_valid(self, form):
        r = super().form_valid(form)
        if self.success_message:
            messages.success(self.request, self.success_message)
        return r


class SchoolTermRelatedExtensibleModel(ExtensibleModel):
    """Add relation to school term."""

    school_term = models.ForeignKey(
        "core.SchoolTerm",
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("Linked school term"),
        blank=True,
        null=True,
    )

    objects = AlekSISBaseManagerWithoutMigrations.from_queryset(SchoolTermRelatedQuerySet)()

    class Meta:
        abstract = True


class SchoolTermRelatedExtensibleForm(ExtensibleForm):
    """Extensible form for school term related data.

    .. warning::
        This doesn't automatically include the field `school_term` in `fields` or `layout`,
        it just sets an initial value.
    """

    def __init__(self, *args, **kwargs):
        from aleksis.core.models import SchoolTerm  # noqa

        if "instance" not in kwargs:
            kwargs["initial"] = {"school_term": SchoolTerm.current}

        super().__init__(*args, **kwargs)


class PublicFilePreferenceMixin(FilePreference):
    """Uploads a file to the public namespace."""

    upload_path = "public"

    def get_upload_path(self):
        return os.path.join(
            self.upload_path, preferences_settings.FILE_PREFERENCE_UPLOAD_DIR, self.identifier()
        )


class RegistryObject:
    """Generic registry to allow registration of subclasses over all apps."""

    _registry: ClassVar[Optional[dict[str, type["RegistryObject"]]]] = None
    name: ClassVar[str] = ""

    def __init_subclass__(cls):
        if getattr(cls, "_registry", None) is None:
            cls._registry = {}
        else:
            if not cls.name:
                cls.name = cls.__name__
            cls._register()

    @classmethod
    def _register(cls: type["RegistryObject"]):
        if cls.name and cls.name not in cls._registry:
            cls._registry[cls.name] = cls

    @classproperty
    def registered_objects_dict(cls) -> dict[str, type["RegistryObject"]]:
        """Get dict of registered objects."""
        return cls._registry

    @classproperty
    def registered_objects_list(cls) -> list[type["RegistryObject"]]:
        """Get list of registered objects."""
        return list(cls._registry.values())

    @classmethod
    def get_object_by_name(cls, name: str) -> Optional[type["RegistryObject"]]:
        """Get registered object by name."""
        return cls.registered_objects_dict.get(name)


class ObjectAuthenticator(RegistryObject):
    def authenticate(self, request, obj):
        raise NotImplementedError()


class CalendarEventMixin(RegistryObject):
    """Mixin for calendar feeds.

    This mixin can be used to create calendar feeds for objects. It can be used
    by adding it to a model or another object. The basic attributes of the calendar
    can be set by either setting the attributes of the class or by implementing
    the corresponding class methods. Please notice that the class methods are
    overriding the attributes. The following attributes are mandatory:

    - name: Unique name for the calendar feed
    - verbose_name: Shown name of the feed

    The respective class methods have a `get_` prefix and are called without any arguments.
    There are also some more attributes. Please refer to the class signature for more
    information.

    The list of objects used to create the calendar feed have to be provided by
    the method `get_objects` class method. It's mandatory to implement this method.

    To provide the data for the events, a certain set of class methods can be implemented.
    The following iCal attributes are supported:

    guid, title, description, link, class, created, updateddate, start_datetime, end_datetime,
    location, geolocation, transparency, organizer, attendee, rrule, rdate, exdate, valarm, status

    Additionally, the color attribute is supported. The color has to be an RGB
    color in the format #ffffff.

    To deploy extra meta data for AlekSIS' own calendar frontend, you can add a
    dictionary for the meta attribute.

    To implement a method for a certain attribute, the name of the method has to be
    `value_<your_attribute>`. For example, to implement the `title` attribute, the
    method `value_title` has to be implemented. The method has to return the value
    for the attribute. The method is called with the reference object as argument.
    """

    name: str = ""  # Unique name for the calendar feed
    verbose_name: str = ""  # Shown name of the feed
    link: str = ""  # Link for the feed, optional
    description: str = ""  # Description of the feed, optional
    color: str = "#222222"  # Color of the feed, optional
    permission_required: str = ""

    @classmethod
    def get_verbose_name(cls, request: Optional[HttpRequest] = None) -> str:
        """Return the verbose name of the calendar feed."""
        return cls.verbose_name

    @classmethod
    def get_link(cls, request: Optional[HttpRequest] = None) -> str:
        """Return the link of the calendar feed."""
        return cls.link

    @classmethod
    def get_description(cls, request: Optional[HttpRequest] = None) -> str:
        """Return the description of the calendar feed."""
        return cls.description

    @classmethod
    def get_language(cls, request: Optional[HttpRequest] = None) -> str:
        """Return the language of the calendar feed."""
        if request:
            return request.LANGUAGE_CODE
        return settings.LANGUAGE_CODE

    @classmethod
    def get_color(cls, request: Optional[HttpRequest] = None) -> str:
        """Return the color of the calendar feed.

        The color has to be an RGB color in the format #ffffff.
        """
        return cls.color

    @classmethod
    def get_enabled(cls, request: HttpRequest | None = None) -> bool:
        """Return whether the calendar is visible in the frontend."""
        if cls.permission_required and request:
            return request.user.has_perm(cls.permission_required)
        return True

    @classmethod
    def create_event(
        cls,
        reference_object: Any,
        feed: ExtendedICal20Feed,
        request: Optional[HttpRequest] = None,
        params: Optional[dict[str, any]] = None,
    ) -> dict[str, Any]:
        """Create an event for the given reference object and add it to the feed."""
        values = {}
        for field in cls.get_event_field_names():
            field_value = cls.get_event_field_value(
                reference_object, field, request=request, params=params
            )
            if field_value is not None:
                values[field] = field_value
        feed.add_item(**values)
        return values

    @classmethod
    def start_feed(
        cls, request: Optional[HttpRequest] = None, params: Optional[dict[str, any]] = None
    ) -> ExtendedICal20Feed:
        """Start the feed and return it."""
        feed = ExtendedICal20Feed(
            title=cls.get_verbose_name(request=request),
            link=cls.get_link(request=request),
            description=cls.get_description(request=request),
            language=cls.get_language(request=request),
            color=cls.get_color(request=request),
        )
        return feed

    @classmethod
    def get_objects(
        cls, request: Optional[HttpRequest] = None, params: Optional[dict[str, any]] = None
    ) -> Iterable:
        """Return the objects to create the calendar feed for."""
        raise NotImplementedError

    @classmethod
    def create_feed(
        cls, request: Optional[HttpRequest] = None, params: Optional[dict[str, any]] = None
    ) -> ExtendedICal20Feed:
        """Create the calendar feed with all events."""
        feed = cls.start_feed(request=request, params=params)

        for reference_object in cls.get_objects(request=request, params=params):
            cls.create_event(reference_object, feed, request=request, params=params)

        return feed

    @classmethod
    def get_calendar_object(
        cls, request: Optional[HttpRequest] = None, params: Optional[dict[str, any]] = None
    ) -> Calendar:
        """Return the calendar object."""
        feed = cls.create_feed(request=request, params=params)
        return feed.get_calendar_object()

    @classmethod
    def get_single_events(
        cls,
        start: datetime | None = None,
        end: datetime | None = None,
        request: HttpRequest | None = None,
        params: dict[str, any] | None = None,
        with_reference_object: bool = False,
    ):
        """Get single events for this calendar feed."""
        feed = cls.create_feed(request=request, params=params)
        return feed.get_single_events(start, end, with_reference_object=with_reference_object)

    @classmethod
    def get_event_field_names(cls) -> list[str]:
        """Return the names of the fields to be used for the feed."""
        return [field_map[0] for field_map in ITEM_ELEMENT_FIELD_MAP]

    @classmethod
    def get_event_field_value(
        cls,
        reference_object,
        field_name: str,
        request: HttpRequest | None = None,
        params: dict[str, Any] | None = None,
    ) -> any:
        """Return the value for the given field name."""
        method_name = f"value_{field_name}"
        if hasattr(cls, method_name) and callable(getattr(cls, method_name)):
            return getattr(cls, method_name)(reference_object, request=request)
        return None

    @classmethod
    def value_link(cls, reference_object, request: HttpRequest | None = None) -> str:
        return ""

    @classmethod
    def value_color(cls, reference_object, request: HttpRequest | None = None) -> str:
        return cls.get_color(request=request)

    @classproperty
    def valid_feed(cls) -> bool:
        """Return if the feed is valid."""
        return cls.name != cls.__name__

    @classproperty
    def valid_feeds(cls):
        """Return a list of valid feeds."""
        return [feed for feed in cls.registered_objects_list if feed.valid_feed]

    @classproperty
    def valid_feed_names(cls) -> list[str]:
        """Return a list of valid feed names."""
        return [feed.name for feed in cls.valid_feeds]

    @classmethod
    def get_object_by_name(cls, name):
        return cls.registered_objects_dict.get(name)

    @classmethod
    def get_activated(cls, person: "Person") -> bool:
        return cls.name in person.preferences["calendar__activated_calendars"]

    @classmethod
    def get_enabled_feeds(cls, request: HttpRequest | None = None):
        return [feed for feed in cls.valid_feeds if feed.get_enabled(request)]



import json

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.db.models import Model

import graphene
from django_filters.filterset import FilterSet, filterset_factory
from graphene_django import DjangoListField, DjangoObjectType

from ..util.core_helpers import queryset_rules_filter


class RulesObjectType(DjangoObjectType):
    class Meta:
        abstract = True

    @classmethod
    def get_queryset(cls, queryset, info, perm):
        q = super().get_queryset(queryset, info)

        return queryset_rules_filter(info.context, q, perm)


class FieldFileType(graphene.ObjectType):
    url = graphene.String()
    absolute_url = graphene.String()

    def resolve_url(root, info, **kwargs):
        return root.url if root else ""

    def resolve_absolute_url(root, info, **kwargs):
        return info.context.build_absolute_uri(root.url) if root else ""


class DeleteMutation(graphene.Mutation):
    """Mutation to delete an object."""

    klass: Model = None
    permission_required: str = ""
    ok = graphene.Boolean()

    class Arguments:
        id = graphene.ID()  # noqa

    @classmethod
    def mutate(cls, root, info, **kwargs):
        obj = cls.klass.objects.get(pk=kwargs["id"])
        if info.context.user.has_perm(cls.permission_required, obj):
            obj.delete()
            return cls(ok=True)
        else:
            raise PermissionDenied()


class PermissionsTypeMixin:
    """Mixin for adding permissions to a Graphene type.

    To configure the names for the permissions or to do
    different permission checking, override the respective
    methods `resolve_can_edit` and `resolve_can_delete`
    """

    can_edit = graphene.Boolean()
    can_delete = graphene.Boolean()

    @staticmethod
    def resolve_can_edit(root: Model, info, **kwargs):
        content_type = ContentType.objects.get_for_model(root)
        perm = f"{content_type.app_label}.edit_{content_type.model}_rule"
        return info.context.user.has_perm(perm, root)

    @staticmethod
    def resolve_can_delete(root: Model, info, **kwargs):
        content_type = ContentType.objects.get_for_model(root)
        perm = f"{content_type.app_label}.delete_{content_type.model}_rule"
        return info.context.user.has_perm(perm, root)


class OptimisticResponseTypeMixin:
    """Mixin for using OptimisticResponse in the frontend.

    In the frontend, Apollo can be configured to show an optimistic
    response defined by the developer although the actual request
    isn't yet finished. This way the frontend can directly update
    and replace the optimistc data with the actual data from the
    backend when the request is finished.

    To know whether a response is optimistic or real, the field
    `is_optimistic` can be set to true in the frontend.
    """

    is_optimistic = graphene.Boolean(default_value=False)


class PermissionBatchPatchMixin:
    class Meta:
        login_required = True

    @classmethod
    def check_permissions(cls, root, info, input):  # noqa
        if info.context.user.has_perms(cls._meta.permissions, root):
            return

        raise PermissionDenied()


class PermissionBatchDeleteMixin:
    class Meta:
        login_required = True

    @classmethod
    def check_permissions(cls, root, info, input):  # noqa
        if info.context.user.has_perms(cls._meta.permissions, root):
            return

        raise PermissionDenied()


class PermissionPatchMixin:
    class Meta:
        login_required = True

    @classmethod
    def check_permissions(cls, root, info, input, id, obj):  # noqa
        if info.context.user.has_perms(cls._meta.permissions, root):
            return

        raise PermissionDenied()


class DjangoFilterMixin:
    """Filters a queryset with django filter."""

    @classmethod
    def get_filterset(cls):
        meta = getattr(cls, "_meta", None)

        if not meta:
            raise NotImplementedError(f"{cls.__name__} must implement class Meta for filtering.")

        if hasattr(meta, "filterset_class"):
            filterset = meta.filterset_class
            if filterset is not None:
                return filterset

        model: Model = meta.model
        fields = getattr(meta, "filter_fields", None)

        if not model:
            raise NotImplementedError(f"{cls.__name__} must supply a model via the Meta class")

        if not fields:
            # Django filter doesn't allow to filter without explicit fields
            raise NotImplementedError(
                f"{cls.__name__}.Meta must contain filter_fields or a filterset_class"
            )

        fs = filterset_factory(model=model, fields=fields)

        return fs

    @classmethod
    def filter(cls, filters, queryset):  # noqa
        filterset_class = cls.get_filterset()
        filterset: FilterSet = filterset_class(filters, queryset)
        return filterset.qs


class FilterOrderList(DjangoListField):
    """Generic filterable Field for lists of django models.

    After the models are filtered, they can be filtered again (e.g.
    for permissions using the get_queryset method inside the
    DjangoObjectType subclass.
    """

    def __init__(self, _type, *args, **kwargs):
        kwargs.update(order_by=graphene.List(graphene.String))
        kwargs.update(filters=graphene.JSONString())
        super().__init__(_type, *args, **kwargs)

    @staticmethod
    def list_resolver(
        django_object_type,
        resolver,
        default_manager,
        root,
        info,
        order_by=None,
        filters=None,
        **args,
    ):
        qs = DjangoListField.list_resolver(
            django_object_type, resolver, default_manager, root, info, **args
        )

        if filters is not None:
            if isinstance(filters, str):
                filters = json.loads(filters)

            if isinstance(filters, dict) and len(filters.keys()) > 0:
                for f_key, f_value in filters.items():
                    if isinstance(f_value, list):
                        filters[f_key] = ",".join(map(str, f_value))

                qs = django_object_type.filter(filters, qs)

        if order_by is not None:
            if isinstance(order_by, str):
                order_by = [order_by]

            qs = qs.order_by(*order_by)

        print(f"{filters=}")

        return qs



# Generated by Django 3.2.11 on 2022-01-08 13:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0030_user_attributes'),
    ]

    operations = [
        migrations.AddField(
            model_name='oauthapplication',
            name='icon',
            field=models.ImageField(blank=True, help_text='This image will be shown as icon in the authorization flow. It should be squared.', null=True, upload_to='', verbose_name='Icon'),
        ),
    ]



"""Custom registries for some preference containers."""

from dynamic_preferences.registries import (
    PerInstancePreferenceRegistry,
    global_preferences_registry,
)


class PersonPreferenceRegistry(PerInstancePreferenceRegistry):
    """Registry for preferences valid for a person."""

    pass


class GroupPreferenceRegistry(PerInstancePreferenceRegistry):
    """Registry for preferences valid for members of a group."""

    pass


site_preferences_registry = global_preferences_registry
person_preferences_registry = PersonPreferenceRegistry()
group_preferences_registry = GroupPreferenceRegistry()



from django.core.exceptions import PermissionDenied

from graphene_django import DjangoObjectType
from guardian.shortcuts import get_objects_for_user

from ..models import Group, Person
from ..util.core_helpers import has_person


class GroupType(DjangoObjectType):
    class Meta:
        model = Group
        fields = [
            "id",
            "school_term",
            "name",
            "short_name",
            "members",
            "owners",
            "child_groups",
            "parent_groups",
            "group_type",
            "photo",
            "avatar",
        ]

    @staticmethod
    def resolve_parent_groups(root, info, **kwargs):
        return get_objects_for_user(info.context.user, "core.view_group", root.parent_groups.all())

    @staticmethod
    def resolve_child_groups(root, info, **kwargs):
        return get_objects_for_user(info.context.user, "core.view_group", root.child_groups.all())

    @staticmethod
    def resolve_members(root, info, **kwargs):
        persons = get_objects_for_user(info.context.user, "core.view_person", root.members.all())
        if has_person(info.context.user) and [
            m for m in root.members.all() if m.pk == info.context.user.person.pk
        ]:
            persons = (persons | Person.objects.filter(pk=info.context.user.person.pk)).distinct()
        return persons

    @staticmethod
    def resolve_owners(root, info, **kwargs):
        persons = get_objects_for_user(info.context.user, "core.view_person", root.owners.all())
        if has_person(info.context.user) and [
            o for o in root.owners.all() if o.pk == info.context.user.person.pk
        ]:
            persons = (persons | Person.objects.filter(pk=info.context.user.person.pk)).distinct()
        return persons

    @staticmethod
    def resolve_group_type(root, info, **kwargs):
        if info.context.user.has_perm("core.view_grouptype_rule", root.group_type):
            return root.group_type
        raise PermissionDenied()



# Generated by Django 3.0.7 on 2020-06-28 11:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0002_school_term'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='person',
            name='photo_cropping',
        ),
        migrations.AlterField(
            model_name='person',
            name='photo',
            field=models.ImageField(blank=True, null=True, upload_to='', verbose_name='Photo'),
        ),
    ]



import graphene
from graphene_django import DjangoObjectType

from aleksis.core.models import OAuthAccessToken, OAuthApplication

from .base import FieldFileType


class OAuthScope(graphene.ObjectType):
    name = graphene.String()
    description = graphene.String()


class OAuthApplicationType(DjangoObjectType):
    icon = graphene.Field(FieldFileType)

    class Meta:
        model = OAuthApplication
        fields = ["id", "name", "icon"]


class OAuthAccessTokenType(DjangoObjectType):
    scopes = graphene.List(OAuthScope)

    @staticmethod
    def resolve_scopes(root: OAuthAccessToken, info, **kwargs):
        return [OAuthScope(name=key, description=value) for key, value in root.scopes.items()]

    class Meta:
        model = OAuthAccessToken
        fields = ["id", "application", "expires", "created", "updated"]


class OAuthBatchRevokeTokenMutation(graphene.Mutation):
    class Arguments:
        ids = graphene.List(graphene.ID)

    ok = graphene.Boolean()

    @staticmethod
    def mutate(root, info, ids):
        OAuthAccessToken.objects.filters(pk__in=ids, user=info.context.user).delete()
        return len(ids)



import base64
import os
import subprocess  # noqa
from datetime import timedelta
from tempfile import TemporaryDirectory
from typing import Callable, Optional, Tuple, Union
from urllib.parse import urljoin

from django.conf import settings
from django.core.files import File
from django.core.files.base import ContentFile
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import get_language
from django.utils.translation import gettext as _

from celery.result import AsyncResult
from celery_progress.backend import ProgressRecorder
from selenium import webdriver

from aleksis.core.celery import app
from aleksis.core.models import PDFFile
from aleksis.core.util.celery_progress import recorded_task, render_progress_page
from aleksis.core.util.core_helpers import has_person, process_custom_context_processors


def _generate_pdf_with_chromium(temp_dir, pdf_path, html_url, lang):
    """Generate a PDF file from a HTML file."""
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument("--kiosk-printing")
    chrome_options.add_argument("--headless")
    chrome_options.add_argument("--no-sandbox")
    chrome_options.add_argument("--disable-gpu")
    chrome_options.add_argument("--disable-dev-shm-usage")
    chrome_options.add_argument("--disable-setuid-sandbox")
    chrome_options.add_argument("--dbus-stub")
    chrome_options.add_argument("--temp-profile")
    chrome_options.add_argument(f"--lang={lang}")

    driver = webdriver.Chrome(options=chrome_options)
    driver.get(html_url)
    pdf = driver.execute_cdp_cmd(
        "Page.printToPDF", {"printBackground": True, "preferCSSPageSize": True}
    )
    driver.close()
    with open(pdf_path, "wb") as f:
        f.write(base64.b64decode(pdf["data"]))


@recorded_task
def generate_pdf(
    file_pk: int, html_url: str, recorder: ProgressRecorder, lang: Optional[str] = None
):
    """Generate a PDF file by rendering the HTML code using a headless Chromium."""
    file_object = get_object_or_404(PDFFile, pk=file_pk)

    recorder.set_progress(0, 1)

    # Open a temporary directory
    with TemporaryDirectory() as temp_dir:
        pdf_path = os.path.join(temp_dir, "print.pdf")
        lang = lang or get_language()

        _generate_pdf_with_chromium(temp_dir, pdf_path, html_url, lang)

        # Upload PDF file to media storage
        with open(pdf_path, "rb") as f:
            file_object.file.save("print.pdf", File(f))
            file_object.save()

    recorder.set_progress(1, 1)


def process_context_for_pdf(context: Optional[dict] = None, request: Optional[HttpRequest] = None):
    context = context or {}
    if not request:
        processed_context = process_custom_context_processors(
            settings.NON_REQUEST_CONTEXT_PROCESSORS
        )
        processed_context.update(context)
    else:
        processed_context = context
    return processed_context


def generate_pdf_from_html(
    html: str, request: Optional[HttpRequest] = None, file_object: Optional[PDFFile] = None
) -> Tuple[PDFFile, AsyncResult]:
    """Start a PDF generation task and return the matching file object and Celery result."""
    html_file = ContentFile(html.encode(), name="source.html")

    # In some cases, the file object is already created (to get a redirect URL for the PDF)
    if not file_object:
        file_object = PDFFile.objects.create()
    if request and has_person(request):
        file_object.person = request.user.person
    file_object.html_file = html_file
    file_object.save()

    # As this method may be run in background and there is no request available,
    # we have to use a predefined URL from settings then
    if request:
        html_url = request.build_absolute_uri(file_object.html_file.url)
    else:
        html_url = urljoin(settings.BASE_URL, file_object.html_file.url)

    result = generate_pdf.delay(file_object.pk, html_url, lang=get_language())

    return file_object, result


def generate_pdf_from_template(
    template_name: str,
    context: Optional[dict] = None,
    request: Optional[HttpRequest] = None,
    render_method: Optional[Callable] = None,
    file_object: Optional[PDFFile] = None,
) -> Tuple[PDFFile, AsyncResult]:
    """Start a PDF generation task and return the matching file object and Celery result."""
    processed_context = process_context_for_pdf(context, request)

    if render_method:
        html_template = render_method(processed_context, request)
    else:
        html_template = render_to_string(template_name, processed_context, request)

    return generate_pdf_from_html(html_template, request, file_object=file_object)


def render_pdf(
    request: Union[HttpRequest, None], template_name: str, context: dict = None
) -> HttpResponse:
    """Start PDF generation and show progress page.

    The progress page will redirect to the PDF after completion.
    """
    if not context:
        context = {}

    file_object, result = generate_pdf_from_template(template_name, context, request)

    redirect_url = f"/pdfs/{file_object.pk}"

    return render_progress_page(
        request,
        result,
        title=_("Progress: Generate PDF file"),
        progress_title=_("Generating PDF file …"),
        success_message=_("The PDF file has been generated successfully."),
        error_message=_("There was a problem while generating the PDF file."),
        redirect_on_success_url=redirect_url,
        back_url=context.get("back_url", reverse("index")),
        button_title=_("Download PDF"),
        button_url=redirect_url,
        button_icon="mdi-file-pdf-box",
    )


def clean_up_expired_pdf_files() -> None:
    """Clean up expired PDF files."""
    PDFFile.objects.filter(expires_at__lt=timezone.now()).delete()


@app.task(run_every=timedelta(days=1))
def clean_up_expired_pdf_files_task() -> None:
    """Clean up expired PDF files."""
    return clean_up_expired_pdf_files()



from django.db import migrations


def forwards_func(apps, schema_editor):
    SitePreferenceModel = apps.get_model("core", "SitePreferenceModel")
    GlobalPreferenceModel = apps.get_model("dynamic_preferences", "GlobalPreferenceModel")
    db_alias = schema_editor.connection.alias
    for pref in SitePreferenceModel.objects.using(db_alias):
        GlobalPreferenceModel.objects.update_or_create(section=pref.section, name=pref.name, defaults=dict(raw_value=pref.raw_value))


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0057_drop_otp_yubikey"),
        ("dynamic_preferences", "0006_auto_20191001_2236")
    ]

    operations = [
        migrations.RunPython(forwards_func),
    ]



from django.conf import settings
from django.utils import translation

import graphene

from ..models import CustomMenu
from ..util.core_helpers import get_site_preferences
from ..util.frontend_helpers import get_language_cookie
from .custom_menu import CustomMenuType
from .site_preferences import SitePreferencesType


class LanguageType(graphene.ObjectType):
    code = graphene.String(required=True)
    name = graphene.String(required=True)
    name_local = graphene.String(required=True)
    name_translated = graphene.String(required=True)
    bidi = graphene.Boolean(required=True)
    cookie = graphene.String(required=True)


class SystemPropertiesType(graphene.ObjectType):
    current_language = graphene.String(required=True)
    default_language = graphene.Field(LanguageType)
    available_languages = graphene.List(LanguageType)
    site_preferences = graphene.Field(SitePreferencesType)
    custom_menu_by_name = graphene.Field(CustomMenuType)

    def resolve_current_language(parent, info, **kwargs):
        return info.context.LANGUAGE_CODE

    @staticmethod
    def resolve_default_language(root, info, **kwargs):
        code = settings.LANGUAGE_CODE
        return translation.get_language_info(code) | {"cookie": get_language_cookie(code)}

    def resolve_available_languages(parent, info, **kwargs):
        return [
            translation.get_language_info(code) | {"cookie": get_language_cookie(code)}
            for code, name in settings.LANGUAGES
        ]

    def resolve_site_preferences(root, info, **kwargs):
        return get_site_preferences()

    def resolve_custom_menu_by_name(root, info, name, **kwargs):
        return CustomMenu.get_default(name)



# Generated by Django 3.2.11 on 2022-01-27 18:52

from email.headerregistry import Group
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0034_invite_permission'),
    ]

    def _migrate_preference_models_to_unique(apps, schema_editor):
        SitePreferenceModel = apps.get_model("core", "SitePreferenceModel")
        PersonPreferenceModel = apps.get_model("core", "PersonPreferenceModel")
        GroupPreferenceModel = apps.get_model("core", "GroupPreferenceModel")

        db_alias = schema_editor.connection.alias

        models = [GroupPreferenceModel, PersonPreferenceModel, SitePreferenceModel]

        for model in models:
            duplicates = {}
            for obj in model.objects.using(db_alias).all():
                key = f"{obj.section}__{obj.name}"
                duplicates.setdefault(key, [])
                duplicates[key].append(obj)

            for key, objs in duplicates.items():
                if len(objs) > 1:
                    found = False
                    for obj in objs:
                        if obj.value == obj.preference.default or found:
                            obj.delete()
                        else:
                            found = True


    operations = [
        migrations.RunPython(_migrate_preference_models_to_unique),
        migrations.AlterUniqueTogether(
            name='grouppreferencemodel',
            unique_together={('section', 'name')},
        ),
        migrations.AlterUniqueTogether(
            name='personpreferencemodel',
            unique_together={('section', 'name')},
        ),
        migrations.AlterUniqueTogether(
            name='sitepreferencemodel',
            unique_together={('section', 'name')},
        ),
    ]



# Generated by Django 4.2.4 on 2023-08-16 20:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0052_site_related_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="FreeBusy",
            fields=[
                (
                    "calendarevent_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.calendarevent",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.calendarevent",),
        ),
    ]
